# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

from jaxley.utils.cell_utils import local_index_of_loc

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

import numpy as np
import pytest

import jaxley as jx
from jaxley.connect import (
    connect,
    connectivity_matrix_connect,
    fully_connect,
    sparse_connect,
)
from jaxley.synapses import IonotropicSynapse, TestSynapse


def test_connect(SimpleBranch, SimpleCell, SimpleNet):
    branch = SimpleBranch(4)
    cell = SimpleCell(3, 4)
    net1 = SimpleNet(4, 3, 8)
    net2 = SimpleNet(4, 3, 8)

    cell1_net1 = net1[0, 0, 0]
    cell2_net1 = net1[1, 0, 0]
    cell1_net2 = net2[0, 0, 0]

    ### test connect single
    # test raise if not part of net
    with pytest.raises(AssertionError):
        connect(cell[0, 0], cell[0, 1], TestSynapse())  # should raise
    with pytest.raises(AssertionError):
        connect(branch[0], branch[1], TestSynapse())  # should raise
    with pytest.raises(AssertionError):
        connect(cell[0, 0], branch[0], TestSynapse())  # should raise

    # test raise if not part of same net
    connect(cell1_net1, cell2_net1, TestSynapse())
    with pytest.raises(AssertionError):
        connect(cell1_net1, cell1_net2, TestSynapse())  # should raise

    ### test connect multiple
    # test connect multiple with single synapse
    connect(net2[1, 0], net2[2, 0], TestSynapse())

    # test after all connections are made, to catch "overwritten" connections
    get_comps = lambda locs: [
        local_index_of_loc(loc, 0, net2.ncomp_per_branch) for loc in locs
    ]

    # check if all connections are made correctly
    first_set_edges = net2.edges.iloc[:8]
    nodes = net2.nodes
    cols = ["pre_index", "post_index"]
    comp_inds = nodes.loc[first_set_edges[cols].to_numpy().flatten()]
    branch_inds = comp_inds["global_branch_index"].to_numpy().reshape(-1, 2)
    cell_inds = comp_inds["global_cell_index"].to_numpy().reshape(-1, 2)
    assert np.all(branch_inds == (3, 6))
    assert (cell_inds == (1, 2)).all()
    assert (
        get_comps(first_set_edges["pre_locs"])
        == get_comps(first_set_edges["post_locs"])
        == [0, 1, 2, 3, 4, 5, 6, 7]
    )
    assert (first_set_edges["type"] == "TestSynapse").all()


def test_fully_connect(SimpleNet):
    net1 = SimpleNet(16, 3, 8)
    for i in range(4):
        fully_connect(net1[i], net1[4:8], TestSynapse())

    fully_connect(net1[8:12], net1[12:16], TestSynapse())

    assert all(
        net1.nodes.loc[net1.edges.post_index, "global_comp_index"]
        == [
            96,
            120,
            144,
            168,
            96,
            120,
            144,
            168,
            96,
            120,
            144,
            168,
            96,
            120,
            144,
            168,
            288,
            312,
            336,
            360,
            288,
            312,
            336,
            360,
            288,
            312,
            336,
            360,
            288,
            312,
            336,
            360,
        ]
    )

    # Test with random post synaptic compartment selection
    net2 = SimpleNet(16, 3, 8)
    _ = np.random.seed(0)
    for i in range(4):
        fully_connect(net2[i], net2[4:8], TestSynapse(), random_post_comp=True)

    fully_connect(net2[8:12], net2[12:16], TestSynapse(), random_post_comp=True)

    assert all(
        net2.nodes.loc[net2.edges.post_index, "global_comp_index"]
        == [
            108,
            135,
            165,
            168,
            99,
            123,
            151,
            177,
            115,
            141,
            162,
            172,
            119,
            126,
            156,
            169,
            294,
            329,
            345,
            379,
            295,
            317,
            356,
            365,
            311,
            325,
            355,
            375,
            302,
            320,
            352,
            375,
        ]
    )


def test_sparse_connect(SimpleNet):
    net1 = SimpleNet(4 * 4, 4, 4)
    _ = np.random.seed(0)

    for i in range(4):
        sparse_connect(net1[i], net1[4:8], TestSynapse(), p=0.1)

    sparse_connect(net1[8:12], net1[12:], TestSynapse(), p=0.5)

    assert all(
        net1.nodes.loc[net1.edges.post_index, "global_comp_index"]
        == [112, 112, 240, 192, 208, 208, 208, 224, 208, 192, 240]
    )

    # Test with random post synaptic compartment selection
    net2 = SimpleNet(4 * 4, 4, 4)
    _ = np.random.seed(0)

    for i in range(4):
        sparse_connect(net2[i], net2[4:8], TestSynapse(), p=0.1, random_post_comp=True)

    sparse_connect(net2[8:12], net2[12:], TestSynapse(), p=0.5, random_post_comp=True)

    assert all(
        net2.nodes.loc[net2.edges.post_index, "global_comp_index"]
        == [123, 201, 196, 211, 208, 211, 213, 238, 255, 255]
    )


def test_connectivity_matrix_connect(SimpleNet):
    net = SimpleNet(4 * 4, 3, 8)

    n_by_n_adjacency_matrix = np.array(
        [[0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1], [1, 0, 0, 0]], dtype=bool
    )
    inds_of_connected_cells = np.stack(np.where(n_by_n_adjacency_matrix)).T
    inds_of_connected_cells[:, 1] += 4  # post-syn cell inds start a 4 below

    connectivity_matrix_connect(
        net[:4], net[4:8], TestSynapse(), n_by_n_adjacency_matrix
    )
    assert len(net.edges.index) == 4
    nodes = net.nodes.set_index("global_comp_index")
    cols = ["pre_index", "post_index"]
    comp_inds = nodes.loc[net.edges[cols].to_numpy().flatten()]
    cell_inds = comp_inds["global_cell_index"].to_numpy().reshape(-1, 2)
    assert np.all(cell_inds == inds_of_connected_cells)
    assert all(
        net.nodes.loc[net.edges.post_index, "global_comp_index"] == [120, 144, 168, 96]
    )

    m_by_n_adjacency_matrix = np.array(
        [[0, 1, 1, 0], [0, 0, 1, 1], [0, 0, 0, 1]], dtype=bool
    )
    inds_of_connected_cells = np.stack(np.where(m_by_n_adjacency_matrix)).T

    net = SimpleNet(4 * 4, 3, 8)
    with pytest.raises(AssertionError):
        connectivity_matrix_connect(
            net[:4], net[:4], TestSynapse(), m_by_n_adjacency_matrix
        )  # should raise

    _ = np.random.seed(0)
    connectivity_matrix_connect(
        net[:3], net[:4], TestSynapse(), m_by_n_adjacency_matrix, random_post_comp=True
    )
    assert len(net.edges.index) == 5
    nodes = net.nodes.set_index("global_comp_index")
    cols = ["pre_index", "post_index"]
    comp_inds = nodes.loc[net.edges[cols].to_numpy().flatten()]
    cell_inds = comp_inds["global_cell_index"].to_numpy().reshape(-1, 2)
    assert np.all(cell_inds == inds_of_connected_cells)
    assert all(
        net.nodes.loc[net.edges.post_index, "global_comp_index"] == [36, 63, 69, 72, 75]
    )

    # Test with different cell views
    net = SimpleNet(4 * 4, 3, 8)
    connectivity_matrix_connect(
        net[1:4], net[2:6], TestSynapse(), m_by_n_adjacency_matrix
    )
    assert len(net.edges.index) == 5
    nodes = net.nodes.set_index("global_comp_index")
    cols = ["pre_index", "post_index"]
    comp_inds = nodes.loc[net.edges[cols].to_numpy().flatten()]
    cell_inds = comp_inds["global_cell_index"].to_numpy().reshape(-1, 2)
    # adjust the cell indices based on the views passed
    inds_of_connected_cells[:, 0] += 1
    inds_of_connected_cells[:, 1] += 2
    assert np.all(cell_inds == inds_of_connected_cells)

    # Test with single compartment cells
    comp = jx.Compartment()
    branch = jx.Branch([comp], ncomp=1)
    cell = jx.Cell([branch], parents=[-1])
    net = jx.Network([cell for _ in range(4 * 4)])
    connectivity_matrix_connect(
        net[1:4], net[2:6], TestSynapse(), m_by_n_adjacency_matrix
    )
    assert len(net.edges.index) == 5
    nodes = net.nodes.set_index("global_comp_index")
    cols = ["pre_index", "post_index"]
    comp_inds = nodes.loc[net.edges[cols].to_numpy().flatten()]
    cell_inds = comp_inds["global_cell_index"].to_numpy().reshape(-1, 2)
    assert np.all(cell_inds == inds_of_connected_cells)



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import Dict, List, Union

import jax.numpy as jnp
import networkx as nx

from jaxley.modules.base import to_graph


def distance_direct(
    startpoint: "View",
    endpoints: Union["Branch", "Cell", "View"],
) -> List[float]:
    """Returns the direct distance between a root and other compartments.

    This function uses ``cell.nodes[['x', 'y', 'z']]`` and computes the euclidean
    distance (i.e., the line of sight distance).

    Args:
        startpoint: A single compartment from which to compute the distance.
        endpoints: One or multiple compartments to which to compute the distance to.

    Returns:
        A list of distances.

    Example usage
    ^^^^^^^^^^^^^

    The following computes the direct (line of sight) distance between the
    zero-eth soma compartment and all other compartments. It then saves this distance
    in `cell.nodes["direct_dist_from_soma"]`.

    ::

        from jaxley.morphology import distance_pathwise

        cell.compute_compartment_centers()  # necessary if you modified branch length.
        direct_dists = distance_direct(cell.soma.branch(0).comp(0), cell)
        cell.nodes["direct_dist_from_soma"] = direct_dists
    """
    assert len(startpoint.nodes.index) == 1, "Cannot use multiple root nodes."

    start_xyz = startpoint.nodes[["x", "y", "z"]].to_numpy()[0]
    end_xyz = endpoints.nodes[["x", "y", "z"]].to_numpy()
    return jnp.sqrt(jnp.sum((start_xyz - end_xyz) ** 2, axis=1))


def distance_pathwise(
    startpoint: "View", endpoints: Union["Branch", "Cell", "View"]
) -> List[float]:
    """Returns the pathwise distance between a root and other compartments.

    We use Dijkstra's algorithm to get the path with the lowest
    number of compartments between start and endpoint. It then computes the
    length of that path in micrometers. Note that, for an uncyclic graph, the
    path with the lowest number of compartments between start and endpoint is
    also the path with the lowest length.

    Args:
        startpoint: A single compartment from which to compute the distance.
        endpoints: One or multiple compartments to which to compute the distance to.

    Returns:
        A list of distances.

    Example usage
    ^^^^^^^^^^^^^

    Example 1: The following computes the pathwise distance between the zero-eth soma
    compartment and all other compartments. It then saves this distance in
    `cell.nodes["path_dist_from_soma"]`.

    ::

        from jaxley.morphology import distance_pathwise

        path_dists = distance_pathwise(cell.soma.branch(0).comp(0), cell)
        cell.nodes["path_dist_from_soma"] = path_dists

    Example 2: The following computes the pathwise distance between two compartments.

    ::

        dist = distance_pathwise(cell.branch(8).comp(2), cell.branch(2).comp(0))
    """
    assert len(startpoint.nodes.index) == 1, "Cannot use multiple root nodes."

    root = startpoint.nodes.index[0]
    endpoint_inds = endpoints.nodes.index
    graph = to_graph(startpoint.base)
    graph = nx.to_undirected(graph)

    # Set default for branchpoints.
    for _, data in graph.nodes(data=True):
        data.setdefault("length", 0.0)

    def edge_weight(u: int, v: int, d: Dict) -> float:
        """
        Args:
            u: Start node of an edge.
            v: End node of an edge.
            d: Dictionary of edge attributes (unused because all our attributes are
                not attributes, not edge attributes).

        Returns:
            The pathwise distance between two nodes.
        """
        return (graph.nodes[u]["length"] + graph.nodes[v]["length"]) / 2

    return [
        nx.dijkstra_path_length(graph, root, end, edge_weight) for end in endpoint_inds
    ]



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import Optional

import jax.numpy as jnp

from jaxley.pumps import Pump


class CaNernstReversal(Pump):
    """Compute Calcium reversal from inner and outer concentration of calcium."""

    def __init__(
        self,
        name: Optional[str] = None,
    ):
        self.current_is_in_mA_per_cm2 = True
        super().__init__(name)
        self.channel_constants = {
            "F": 96485.3329,  # C/mol (Faraday's constant)
            "T": 279.45,  # Kelvin (temperature)
            "R": 8.314,  # J/(mol K) (gas constant)
        }
        self.channel_params = {}
        self.channel_states = {"eCa": 0.0, "CaCon_i": 5e-05, "CaCon_e": 2.0}
        # Note that the `self.ion_name` does not matter here, because `compute_current`
        # returns 0.0 (and `self.ion_name` only sets which ion concentration the
        # current should be added to).
        self.ion_name = "CaCon_i"
        self.current_name = f"i_Ca"
        self.META = {"ion": "Ca"}

    def update_states(self, u, dt, voltages, params):
        """Update internal calcium concentration based on calcium current and decay."""
        R, T, F = (
            self.channel_constants["R"],
            self.channel_constants["T"],
            self.channel_constants["F"],
        )
        Cai = u["CaCon_i"]
        Cao = u["CaCon_e"]
        C = R * T / (2 * F) * 1000  # mV
        vCa = C * jnp.log(Cao / Cai)
        return {"eCa": vCa, "CaCon_i": Cai, "CaCon_e": Cao}

    def compute_current(self, u, voltages, params):
        """This dynamics model does not directly contribute to the membrane current."""
        return 0

    def init_state(self, states, voltages, params, delta_t):
        """Initialize the state at fixed point of gate dynamics."""
        return {}



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import os

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

import jax.numpy as jnp
import numpy as np
import pytest

import jaxley as jx
from jaxley.channels import HH


@pytest.mark.parametrize(
    "property", ["radius", "capacitance", "length", "axial_resistivity"]
)
def test_raise_for_heterogenous_modules(SimpleBranch, property):
    branch0 = SimpleBranch(4)
    branch1 = SimpleBranch(4)
    branch1.comp(1).set(property, 1.5)
    cell = jx.Cell([branch0, branch1], parents=[-1, 0])
    with pytest.raises(ValueError):
        cell.branch(1).set_ncomp(2)


def test_raise_for_heterogenous_channel_existence(SimpleBranch):
    branch0 = SimpleBranch(4)
    branch1 = SimpleBranch(4)
    branch1.comp(2).insert(HH())
    cell = jx.Cell([branch0, branch1], parents=[-1, 0])
    with pytest.raises(ValueError):
        cell.branch(1).set_ncomp(2)


def test_raise_for_heterogenous_channel_properties(SimpleBranch):
    branch0 = SimpleBranch(4)
    branch1 = SimpleBranch(4)
    branch1.insert(HH())
    branch1.comp(3).set("HH_gNa", 0.5)
    cell = jx.Cell([branch0, branch1], parents=[-1, 0])
    with pytest.raises(ValueError):
        cell.branch(1).set_ncomp(2)


def test_raise_for_entire_cells(SimpleCell):
    cell = SimpleCell(3, 4)
    with pytest.raises(AssertionError):
        cell.set_ncomp(2)


def test_raise_for_networks(SimpleCell):
    cell1 = SimpleCell(3, 4)
    cell2 = SimpleCell(3, 4)
    net = jx.Network([cell1, cell2])
    with pytest.raises(AssertionError):
        net.cell(0).branch(1).set_ncomp(2)


def test_raise_for_recording(SimpleCell):
    cell = SimpleCell(3, 2)
    cell.branch(0).comp(0).record()
    with pytest.raises(AssertionError):
        cell.branch(1).set_ncomp(2)


def test_raise_for_stimulus(SimpleCell):
    cell = SimpleCell(3, 2)
    cell.branch(0).comp(0).stimulate(0.4 * jnp.ones(100))
    with pytest.raises(AssertionError):
        cell.branch(1).set_ncomp(2)


@pytest.mark.parametrize("new_ncomp", [1, 2, 4, 5, 8])
def test_simulation_accuracy_api_equivalence_init_vs_setncomp_branch(
    SimpleBranch, new_ncomp
):
    """Test whether a module built from scratch matches module built with `set_ncomp()`.

    This makes one branch, whose `ncomp` is not modified, heterogenous.
    """
    branch1 = SimpleBranch(new_ncomp)

    # The second branch is originally instantiated to have 4 ncomp, but is later
    # modified to have `new_ncomp` compartments.
    branch2 = SimpleBranch(4)
    branch2.comp("all").set("length", 10.0)
    total_branch_len = 4 * 10.0

    # Make the total branch length 40 um.
    branch1.comp("all").set("length", total_branch_len / new_ncomp)

    # Adapt ncomp.
    branch2.set_ncomp(new_ncomp)

    for branch in [branch1, branch2]:
        branch.comp(0).stimulate(0.4 * jnp.ones(100))
        branch.comp(new_ncomp - 1).record()

    v1 = jx.integrate(branch1)
    v2 = jx.integrate(branch2)
    max_error = np.max(np.abs(v1 - v2))
    assert max_error < 1e-8, f"Too large voltage deviation, {max_error} > 1e-8"


@pytest.mark.parametrize("new_ncomp", [1, 2, 4, 5, 8])
def test_simulation_accuracy_api_equivalence_init_vs_setncomp_cell(
    SimpleBranch, new_ncomp
):
    """Test whether a module built from scratch matches module built with `set_ncomp()`."""
    branch1 = SimpleBranch(new_ncomp)

    # The second branch is originally instantiated to have 4 ncomp, but is later
    # modified to have `new_ncomp` compartments.
    branch2 = SimpleBranch(4)
    branch2.comp("all").set("length", 10.0)
    total_branch_len = 4 * 10.0

    # Make the total branch length 20 um.
    branch1.comp("all").set("length", total_branch_len / new_ncomp)
    cell1 = jx.Cell(branch1, parents=[-1, 0])
    cell2 = jx.Cell(branch2, parents=[-1, 0])

    # Adapt ncomp.
    for b in range(2):
        cell2.branch(b).set_ncomp(new_ncomp)

    for cell in [cell1, cell2]:
        cell.branch(0).comp(0).stimulate(0.4 * jnp.ones(100))
        cell.branch(1).comp(new_ncomp - 1).record()

    v1 = jx.integrate(cell1)
    v2 = jx.integrate(cell2)
    max_error = np.max(np.abs(v1 - v2))
    assert max_error < 1e-8, f"Too large voltage deviation, {max_error} > 1e-8"


@pytest.mark.parametrize("new_ncomp", [1, 4, 5])
@pytest.mark.parametrize(
    "file", ["morph_ca1_n120_250.swc", "morph_3_types_single_point_soma.swc"]
)
def test_api_equivalence_swc_lengths_and_radiuses(SimpleMorphCell, new_ncomp, file):
    """Test if the radiuses and lengths of an SWC morph are reconstructed correctly."""
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", file)

    cell1 = SimpleMorphCell(fname, ncomp=new_ncomp)
    cell2 = SimpleMorphCell(fname, ncomp=1)

    for b in range(cell2.total_nbranches):
        cell2.branch(b).set_ncomp(new_ncomp)

    for property_name in ["radius", "length"]:
        cell1_vals = cell1.nodes[property_name].to_numpy()
        cell2_vals = cell2.nodes[property_name].to_numpy()
        assert np.allclose(
            cell1_vals, cell2_vals
        ), f"Too large difference in {property_name}"


@pytest.mark.parametrize("new_ncomp", [1, 2, 4, 5, 8])
@pytest.mark.parametrize("file", ["morph_ca1_n120_250.swc"])
def test_simulation_accuracy_swc_init_vs_set_ncomp(SimpleMorphCell, new_ncomp, file):
    """Test whether an SWC initially built with 4 ncomp works after `set_ncomp()`."""
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", file)

    cell1 = SimpleMorphCell(fname, ncomp=new_ncomp)
    cell2 = SimpleMorphCell(fname, ncomp=1)

    for b in range(cell2.total_nbranches):
        cell2.branch(b).set_ncomp(new_ncomp)

    for cell in [cell1, cell2]:
        cell.branch(0).comp(0).stimulate(0.4 * jnp.ones(100))
        cell.branch(0).comp(new_ncomp - 1).record()
        cell.branch(3).comp(0).record()
        cell.branch(5).comp(new_ncomp - 1).record()

    v1 = jx.integrate(cell1, voltage_solver="jax.sparse")
    v2 = jx.integrate(cell2, voltage_solver="jax.sparse")
    max_error = np.max(np.abs(v1 - v2))
    assert max_error < 1e-8, f"Too large voltage deviation, {max_error} > 1e-8"


def test_set_ncomp_raises_for_inhomogeneous(SimpleCell):
    """Test whether groups get updated appropriately after `set_ncomp`."""
    cell = SimpleCell(3, 4)
    cell.branch(0).comp(0).add_to_group("exc")
    with pytest.raises(AssertionError):
        cell.branch(0).set_ncomp(2)


def test_dlambda_rule(SimpleMorphCell):
    """Run `set_ncomp` on every branch with the d-lambda rule.

    If you have to modify this test, make sure to also modify the how-to guide."""
    file = "morph_ca1_n120_250.swc"
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", file)

    cell = SimpleMorphCell(fname, ncomp=1)

    # Reasonable default values for most models.
    frequency = 100.0
    d_lambda = 0.1  # Larger -> more coarse-grained.

    for branch in cell.branches:
        diameter = 2 * branch.nodes["radius"].to_numpy()[0]
        c_m = branch.nodes["capacitance"].to_numpy()[0]
        r_a = branch.nodes["axial_resistivity"].to_numpy()[0]
        l = branch.nodes["length"].to_numpy()[0]

        lambda_f = 1e5 * np.sqrt(diameter / (4 * np.pi * frequency * c_m * r_a))
        ncomp = int((l / (d_lambda * lambda_f) + 0.9) / 2) * 2 + 1
        branch.set_ncomp(ncomp, initialize=False)

    # After the loop, you have to run `cell.initialize()` because we passed
    # `set_ncomp(..., initialize=False)` for speeding up the loop over branches.
    cell.initialize()
    cell[0, 0].record()
    v = jx.integrate(cell, t_max=1.0)
    assert np.invert(np.any(np.isnan(v))), "Found NaN after d_lambda rule."



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import Callable, Dict, List, Optional, Tuple, Union
from warnings import warn

import jax.numpy as jnp
import numpy as np
import pandas as pd

from jaxley.modules.base import Module
from jaxley.modules.compartment import Compartment
from jaxley.utils.cell_utils import compute_children_and_parents
from jaxley.utils.misc_utils import cumsum_leading_zero, deprecated_kwargs


class Branch(Module):
    """A branch made up of one or multiple compartments (without branchpoints).

    This class defines a single branch that can be simulated by itself or
    connected to build a cell. A branch is an unbranched neurite of several
    compartments and can be connected to no, one or more other branches at each end to
    build more intricate cell morphologies (via ``jx.Cell``).
    """

    branch_params: Dict = {}
    branch_states: Dict = {}

    def __init__(
        self,
        compartments: Optional[Union[Compartment, List[Compartment]]] = None,
        ncomp: Optional[int] = None,
    ):
        """
        Args:
            compartments: A single compartment or a list of compartments that make up the
                branch.
            ncomp: Number of segments to divide the branch into. If `compartments` is an
                a single compartment, than the compartment is repeated `ncomp` times to
                create the branch.
        """
        super().__init__()
        assert (
            isinstance(compartments, (Compartment, List)) or compartments is None
        ), "Only Compartment or List[Compartment] is allowed."
        if isinstance(compartments, Compartment):
            assert (
                ncomp is not None
            ), "If `compartments` is not a list then you have to set `ncomp`."
        compartments = Compartment() if compartments is None else compartments
        ncomp = 1 if ncomp is None else ncomp

        if isinstance(compartments, Compartment):
            compartment_list = [compartments] * ncomp
        else:
            compartment_list = compartments

        self.ncomp = len(compartment_list)
        self.ncomp_per_branch = np.asarray([self.ncomp])
        self.total_nbranches = 1
        self.nbranches_per_cell = [1]
        self._cumsum_nbranches = jnp.asarray([0, 1])
        self.cumsum_ncomp = cumsum_leading_zero(self.ncomp_per_branch)

        # Indexing.
        self.nodes = pd.concat([c.nodes for c in compartment_list], ignore_index=True)
        self._append_params_and_states(self.branch_params, self.branch_states)
        self.nodes["global_comp_index"] = np.arange(self.ncomp).tolist()
        self.nodes["global_branch_index"] = [0] * self.ncomp
        self.nodes["global_cell_index"] = [0] * self.ncomp

        # Channels.
        self._gather_channels_from_constituents(compartment_list)

        self.branch_edges = pd.DataFrame(
            dict(parent_branch_index=[], child_branch_index=[])
        )

        # For morphology indexing.
        self._par_inds, self._child_inds, self._child_belongs_to_branchpoint = (
            compute_children_and_parents(self.branch_edges)
        )
        self._internal_node_inds = jnp.arange(self.ncomp)

        # Coordinates.
        self.xyzr = [float("NaN") * np.zeros((2, 4))]
        self.initialize()

    def _init_comp_graph(self):
        """Initialize `._comp_edges`, `._branchpoints`, and `comp_to_index_mapping`.

        It also initializes `_comp_edges_in_view` and `_branchpoints_in_view`."""
        # Compartment edges.
        self._comp_edges = pd.DataFrame().from_dict(
            {
                "source": list(range(self.ncomp - 1)) + list(range(1, self.ncomp)),
                "sink": list(range(1, self.ncomp)) + list(range(self.ncomp - 1)),
                "ordered": list([1] * (self.ncomp - 1)) + list([0] * (self.ncomp - 1)),
            }
        )
        self._comp_edges["type"] = 0
        self._n_nodes = self.ncomp

        # `off_diagonal_inds`.
        sources = np.asarray(self._comp_edges["source"].to_list())
        sinks = np.asarray(self._comp_edges["sink"].to_list())
        self._off_diagonal_inds = jnp.stack([sources, sinks]).astype(int)

    def __len__(self) -> int:
        return self.ncomp



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>
from typing import Any, Dict

import jax
import jax.numpy as jnp
import numpy as np
from jax import Array
from jax.experimental.sparse.linalg import spsolve as jax_spsolve
from jax.lax import fori_loop
from jax.typing import ArrayLike
from tridiax.stone import stone_backsub_lower, stone_triang_upper

from jaxley.solver_gate import exponential_euler


def step_voltage_implicit_with_dhs_solve(
    voltages,
    voltage_terms,
    constant_terms,
    axial_conductances,
    internal_node_inds,
    sinks,
    n_nodes,
    solve_indexer: Dict[str, Any],
    optimize_for_gpu: bool,
    delta_t,
):
    """Return voltage update via compartment-based matrix inverse.

    Combined with an appropriate solve order, this results in `dendritic hierarchical
    scheduling` (DHS, Zhang et al., 2023). The solve order is defined in the graph
    backend of `Jaxley`.

    Args:
        ordered_comp_edges: A list which edges between compartments, each being a
            tuple (child_compartment, parent_compartment). The order of the list
            indicates the solve order.
        map_to_solve_order: An array of indices that permutes diagonal elements into
            the order of the solve. E.g.: `voltages = voltages[mapping_array]`.
        inv_map_to_solve_order: An array of indices that permutes diagonal elements back
            into compartment order. E.g.: `voltages = voltages[inv_mapping_array]`.
        map_to_solve_order_lower_and_upper: An array of indices that permutes
            the concatenation of lowers and uppers into the order of the solve:
            `lowers_and_uppers = lowers_and_uppers[map_to_solve_order_lower_and_upper]`.
        optimize_for_gpu: If True, it does two things: (1) it unrolls the for-loop
            for the triangularization stage. (2) It uses recursive doubling (also
            unrolled) for the backsubstitution stage. Setting this to `True` will
            largely speed up runs on GPU, but it will slow down compilation time and
            run time on CPU.
    """
    axial_conductances = delta_t * axial_conductances

    # Build diagonals.
    diags = delta_t * voltage_terms
    diags = diags.at[internal_node_inds].add(1.0)
    #
    # if-case needed because `.at` does not allow empty inputs, but the input is
    # empty for compartments.
    if len(sinks) > 0:
        diags = diags.at[sinks].add(axial_conductances)

    # Build solve.
    solves = jnp.zeros(n_nodes)
    solves = solves.at[internal_node_inds].set(
        voltages[internal_node_inds] + delta_t * constant_terms[internal_node_inds]
    )

    # Why `n_nodes > 1`? For compartments (or point neurons), we save computation and
    # compile time by skipping the entire solve procedure.
    if len(axial_conductances) > 0:
        # Build lower and upper matrix.
        lowers_and_uppers = -axial_conductances

        # Reorder diagonals and solves.
        diags = diags[solve_indexer["map_to_solve_order"]]
        solves = solves[solve_indexer["map_to_solve_order"]]

        # Reorder the lower and upper values.
        lowers = lowers_and_uppers[solve_indexer["map_to_solve_order_lower"]]
        uppers = lowers_and_uppers[solve_indexer["map_to_solve_order_upper"]]
        ordered_comp_edges = solve_indexer["node_order_grouped"]
        flipped_comp_edges = list(reversed(ordered_comp_edges))

        # Add a spurious compartment that is modified by the masking.
        diags = jnp.concatenate([diags, jnp.asarray([1.0])])
        solves = jnp.concatenate([solves, jnp.asarray([0.0])])
        uppers = jnp.concatenate([uppers, jnp.asarray([0.0])])
        lowers = jnp.concatenate([lowers, jnp.asarray([0.0])])

        # Solve the voltage equations.
        #
        steps = len(flipped_comp_edges)
        if not optimize_for_gpu:
            # Cast from a list to a np.array.
            # `ordered_comp_edges` has shape `(num_levels, num_comps_per_level, 2)`,
            # and `num_comps_per_level=1` for CPU.
            ordered_comp_edges = np.asarray(ordered_comp_edges)
            flipped_comp_edges = np.asarray(flipped_comp_edges)

            # Triangulate.
            steps = len(flipped_comp_edges)
            init = (diags, solves, lowers, uppers, flipped_comp_edges)
            diags, solves, _, _, _ = fori_loop(0, steps, _comp_based_triang, init)

            # Backsubstitute.
            lowers /= diags
            solves /= diags
            diags = jnp.ones_like(solves)
            init = (solves, lowers, ordered_comp_edges)
            solves, _, _ = fori_loop(0, steps, _comp_based_backsub, init)
        else:
            # Triangulate by unrolling the loop of the levels.
            for i in range(steps):
                diags, solves, _, _, _ = _comp_based_triang(
                    i, (diags, solves, lowers, uppers, flipped_comp_edges)
                )

            # Backsubstitute with recursive doubling.
            diags, solves = _comp_based_backsub_recursive_doubling(
                diags, solves, lowers, steps, n_nodes, solve_indexer["parent_lookup"]
            )

        # Remove the spurious compartment. This compartment got modified by masking of
        # compartments in certain levels.
        diags = diags[:-1]
        solves = solves[:-1]

    # Get inverse of the diagonalized matrix.
    solution = solves / diags
    solution = solution[solve_indexer["inv_map_to_solve_order"]]

    return solution


def _comp_based_triang(index, carry):
    """Triangulate the quasi-tridiagonal system compartment by compartment."""
    diags, solves, lowers, uppers, flipped_comp_edges = carry

    # `flipped_comp_edges` has shape `(num_levels, num_comps_per_level, 2)`. We first
    # get the relevant level with `[index]` and then we get all children and parents
    # in the level.
    comp_edge = flipped_comp_edges[index]
    child = comp_edge[:, 0]
    parent = comp_edge[:, 1]

    lower_val = lowers[child]
    upper_val = uppers[child]
    child_diag = diags[child]
    child_solve = solves[child]

    # Factor that the child row has to be multiplied by.
    multiplier = upper_val / child_diag

    # Updates to diagonal and solve
    diags = diags.at[parent].add(-lower_val * multiplier)
    solves = solves.at[parent].add(-child_solve * multiplier)

    return (diags, solves, lowers, uppers, flipped_comp_edges)


def _comp_based_backsub(index, carry):
    """Backsubstitute the quasi-tridiagonal system compartment by compartment."""
    solves, lowers, comp_edges = carry

    # `comp_edges` has shape `(num_levels, num_comps_per_level, 2)`. We first get the
    # relevant level with `[index]` and then we get all children and parents in the
    # level.
    comp_edge = comp_edges[index]
    child = comp_edge[:, 0]
    parent = comp_edge[:, 1]

    # Updates to diagonal and solve
    solves = solves.at[child].add(-solves[parent] * lowers[child])
    return (solves, lowers, comp_edges)


def _comp_based_backsub_recursive_doubling(
    diags: ArrayLike,
    solves: ArrayLike,
    lowers: ArrayLike,
    steps: int,
    n_nodes: int,
    parent_lookup: np.ndarray,
) -> tuple[Array, Array]:
    """Backsubstitute with recursive doubling.

    This function contains a lot of math, so I will describe what is going on here:

    The matrix describes a system like:
    diag[n] * x[n] + lower[n] * x[parent] = solve[n]

    We rephrase this as:
    x[n] = solve[n]/diag[n] - lower[n]/diag[n] * x[parent].

    and we call variables as follows:
    solve/diag => solve_effect
    -lower/diag => lower_effect

    This gives:
    x[n] = solve_effect[n] + lower_effect[n] * x[parent].

    Recursive doubling solves this equation for `x` in log_2(N) steps. How?

    (1) Notice that lower_effect[n]=0, because x[0] has no parent.

    (2) In the first step, recursive doubling substitutes x[parent] into
    every equation. This leads to something like:
    x[n] = solve_effect[n] + lower_effect[n] * (solve_effect[parent] + ...
    ...lower_effect[parent] * x[parent[parent]])

    Abbreviate this as:
    new_solve_effect[n] = solve_effect[n] + lower_effect[n] * solve_effect[parent]
    new_lower_effect[n] = lower_effect[n] + lower_effect[parent]
    x[n] = new_solve_effect[n] + new_lower_effect[n] * x[parent[parent]]
    Importantly, every node n is now a function of its two-step parent.

    (3) In the next step, recursive doubling substitutes x[parent[parent]].
    Since x[parent[parent]] already depends on its own _two-step_ parent,
    every node then depends on its four step parent. This introduces the
    log_2 scaling.

    (4) The algorithm terminates when all `new_lower_effect=0`. This
    naturally happens because `lower_effect[0]=0`, and the recursion
    keeps multiplying new_lower_effect with the `lower_effect[parent]`.
    """
    # Why `lowers = lowers.at[0].set(0.0)`? During triangulation (and the
    # cpu-optimized solver), we never access `lowers[0]`. Its value should
    # be zero (because the zero-eth compartment does not have a `lower`), but
    # it is not for coding convenience in the other solvers. For the recursive
    # doubling solver below, we do use lowers[0], so we set it to the value
    # it should have anyways: 0.
    lowers = lowers.at[0].set(0.0)

    # Rephrase the equations as a recursion.
    # x[n] = solve[n]/diag[n] - lower[n]/diag[n] * x[parent].
    # x[n] = solve_effect[n] + lower_effect[n] * x[parent].
    lower_effect = -lowers / diags
    solve_effect = solves / diags

    step = 1
    while step <= steps:
        # For each node, get its k-step parent, where k=`step`.
        k_step_parent = np.arange(n_nodes + 1)
        for _ in range(step):
            k_step_parent = parent_lookup[k_step_parent]

        # Update.
        solve_effect = lower_effect * solve_effect[k_step_parent] + solve_effect
        lower_effect *= lower_effect[k_step_parent]
        step *= 2

    # We have to return a `diags` because the final solution is computed as
    # `solves/diags` (see `step_voltage_implicit_with_dhs_solve`). For recursive
    # doubling, the solution should just be `solve_effect`, so we define diags as
    # 1.0 so the division has no effect.
    diags = jnp.ones_like(solve_effect)
    solves = solve_effect
    return diags, solves


def step_voltage_implicit_with_jax_spsolve(
    voltages,
    voltage_terms,
    constant_terms,
    axial_conductances,
    internal_node_inds,
    sinks,
    data_inds,
    indices,
    indptr,
    n_nodes,
    delta_t,
):
    axial_conductances = delta_t * axial_conductances

    # Build diagonals.
    diags = delta_t * voltage_terms
    diags = diags.at[internal_node_inds].add(1.0)
    #
    # if-case needed because `.at` does not allow empty inputs, but the input is
    # empty for compartments.
    if len(sinks) > 0:
        diags = diags.at[sinks].add(axial_conductances)

    # Build solve.
    solves = jnp.zeros(n_nodes)
    solves = solves.at[internal_node_inds].set(
        voltages[internal_node_inds] + delta_t * constant_terms[internal_node_inds]
    )

    # Concatenate diagonals and off-diagonals (which are just `-axial_conductances`).
    all_values = jnp.concatenate([diags, -axial_conductances])

    # Solve the voltage equations.
    solution = jax_spsolve(all_values[data_inds], indices, indptr, solves)
    return solution


def step_voltage_explicit(
    voltages: ArrayLike,
    voltage_terms: ArrayLike,
    constant_terms: ArrayLike,
    axial_conductances: ArrayLike,
    sinks,
    sources,
    types,
    n_nodes,
    delta_t: float,
) -> Array:
    """Solve one timestep of branched nerve equations with explicit (forward) Euler."""
    update = _voltage_vectorfield(
        voltages,
        voltage_terms,
        constant_terms,
        axial_conductances,
        sinks,
        sources,
        types,
        n_nodes,
    )
    new_voltates = voltages + delta_t * update
    return new_voltates


def _voltage_vectorfield(
    voltages: ArrayLike,
    voltage_terms: ArrayLike,
    constant_terms: ArrayLike,
    axial_conductances: ArrayLike,
    sinks,
    sources,
    types,
    n_nodes,
) -> Array:
    """Evaluate the vectorfield of the nerve equation."""
    # Membrane current update.
    vecfield = -voltage_terms * voltages + constant_terms

    # Compute branchpoint voltages. The voltages at branchpoints are a weighted
    # average of all neighboring compartments: vBP = (g1 v1 + g2 v2) / (g1 + g2)
    condition = np.isin(types, [3, 4])
    if np.sum(condition) > 0:
        sink_comp_inds = sinks[condition]
        source_comp_inds = sources[condition]
        # Set set the voltages at branchpoints to zero.
        voltages = voltages.at[sink_comp_inds].set(0.0)
        # Compute all terms that influence the branchpoints: g1 v1 + g2 v2
        voltages = voltages.at[sink_comp_inds].add(
            axial_conductances[condition] * voltages[source_comp_inds]
        )
        # Compute the normalizer for the branchpoint: (g1 + g2)
        # Initialize at one such that we can also just divide the compartment voltages
        # by it.
        summed_axials = jnp.ones(n_nodes)
        summed_axials = summed_axials.at[sink_comp_inds].set(0.0)
        summed_axials = summed_axials.at[sink_comp_inds].add(
            axial_conductances[condition]
        )
        voltages = voltages / summed_axials

    # Update compartment voltages with the impact from other compartments and from
    # branchpoints.
    condition = np.isin(types, [0, 1, 2])
    if np.sum(condition) > 0:
        sink_comp_inds = sinks[condition]
        source_comp_inds = sources[condition]
        vecfield = vecfield.at[sinks].add(
            (voltages[sources] - voltages[sinks]) * axial_conductances
        )
    return vecfield


def step_voltage_exponential(
    voltages: ArrayLike,
    voltage_terms: ArrayLike,
    constant_terms: ArrayLike,
    axial_conductances: ArrayLike,
    matrix_exponential,
    internal_node_inds,
    delta_t: float,
) -> Array:
    """Solve one timestep of branched nerve equations with exponential Euler.

    This performs Lie-Trotter splitting for the voltages:

    dv/dt = G * v + g_channel * (E - V)

    v_{t+1} = exp(dt * G) * exp_euler(channel_terms, v_{t})

    I.e., with Lie-Trotter splitting, we first perform an update of the term related to
    the channels, and then update the 'diffusive' term G * v.

    Args:
        voltages: The voltages (including branchpoint voltages).
        voltage_terms: The linear part of the channels terms, i.e., g_channel.
        constant_terms: The constant part of the channel terms, i.e., g_channel * E
        axial_conductances: Unused (because we use the `matrix_exponential`).
        matrix_exponential: The matrix exp(G * dt).
        internal_node_inds: Indexes into all compartments.
        delta_t: Time step.

    Returns:
        The updated voltages (including branchpoint voltages).
    """
    # Update of the channels. This requires a `jnp.where` because: some compartments
    # might not have _any_ channels. In that case, their `voltate_terms = 0`, which
    # causes tau = inf. This if-case solves this.
    tau_is_inf = voltage_terms == 0
    v_tau_not_inf = exponential_euler(
        voltages, delta_t, constant_terms / voltage_terms, 1.0 / voltage_terms
    )
    v_tau_is_inf = voltages + constant_terms * delta_t
    voltages = jnp.where(tau_is_inf, v_tau_is_inf, v_tau_not_inf)

    # Update of the diffusion matrix.
    return voltages.at[internal_node_inds].set(
        matrix_exponential @ voltages[internal_node_inds]
    )


def step_voltage_implicit_with_stone(
    voltages: ArrayLike,
    voltage_terms: ArrayLike,
    constant_terms: ArrayLike,
    axial_conductances: ArrayLike,
    internal_node_inds: ArrayLike,
    n_nodes: int,
    sinks: ArrayLike,
    sources: ArrayLike,
    types: ArrayLike,
    delta_t: float,
):
    """Solve one timestep of branched nerve equations with implicit (backward) Euler."""
    if np.sum(np.isin(types, [1, 2, 3, 4])) > 0:
        raise NotImplementedError(
            f"The stone solver is not implemented for branched morphologies."
        )
    axial_conductances = delta_t * axial_conductances

    # Build diagonals.
    diags = delta_t * voltage_terms
    diags = diags.at[internal_node_inds].add(1.0)
    #
    # if-case needed because `.at` does not allow empty inputs, but the input is
    # empty for compartments.
    if len(sinks) > 0:
        diags = diags.at[sinks].add(axial_conductances)

    lower_inds = sinks > sources
    lowers = -axial_conductances[lower_inds]

    upper_inds = sinks < sources
    uppers = -axial_conductances[upper_inds]

    # Build solve.
    solves = jnp.zeros(n_nodes)
    solves = solves.at[internal_node_inds].set(
        voltages[internal_node_inds] + delta_t * constant_terms[internal_node_inds]
    )

    # Solve the tridiagonal system.
    diags, lowers, solves = stone_triang_upper(lowers, diags, uppers, solves)
    solves = stone_backsub_lower(solves, lowers, diags)

    return solves



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from jaxley.synapses.ionotropic import IonotropicSynapse
from jaxley.synapses.synapse import Synapse
from jaxley.synapses.tanh_conductance import TanhConductanceSynapse
from jaxley.synapses.tanh_rate import TanhRateSynapse
from jaxley.synapses.test import TestSynapse

__all__ = [
    "IonotropicSynapse",
    "Synapse",
    "TanhConductanceSynapse",
    "TanhRateSynapse",
    "TestSynapse",
]



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import json
import os
from copy import deepcopy
from typing import Optional

import pytest

import jaxley as jx
from jaxley.synapses import IonotropicSynapse
from tests.test_regression import generate_regression_report, load_json


@pytest.fixture(scope="session")
def SimpleComp():
    """Fixture for creating or retrieving an already created compartment."""
    comps = {}

    def get_or_build_comp(
        copy: bool = True, force_init: bool = False
    ) -> jx.Compartment:
        """Create or retrieve a compartment.

        Args:
            copy: Whether to return a copy of the compartment. Default is True.
            force_init: Force the init from scratch. Default is False.

        Returns:
            jx.Compartment()."""
        if "comp" not in comps or force_init:
            comps["comp"] = jx.Compartment()
        return deepcopy(comps["comp"]) if copy and not force_init else comps["comp"]

    yield get_or_build_comp
    comps = {}


@pytest.fixture(scope="session")
def SimpleBranch(SimpleComp):
    """Fixture for creating or retrieving an already created branch."""
    branches = {}

    def get_or_build_branch(
        ncomp: int, copy: bool = True, force_init: bool = False
    ) -> jx.Branch:
        """Create or retrieve a branch.

        If a branch with the same number of compartments already exists, it is returned.

        Args:
            ncomp: Number of compartments in the branch.
            copy: Whether to return a copy of the branch. Default is True.
            force_init: Force the init from scratch. Default is False.

        Returns:
            jx.Branch()."""
        if ncomp not in branches or force_init:
            comp = SimpleComp(force_init=force_init)
            branches[ncomp] = jx.Branch([comp] * ncomp)
        return deepcopy(branches[ncomp]) if copy and not force_init else branches[ncomp]

    yield get_or_build_branch
    branches = {}


@pytest.fixture(scope="session")
def SimpleCell(SimpleBranch):
    """Fixture for creating or retrieving an already created cell."""
    cells = {}

    def get_or_build_cell(
        nbranches: int, ncomp: int, copy: bool = True, force_init: bool = False
    ) -> jx.Cell:
        """Create or retrieve a cell.

        If a cell with the same number of branches and compartments already exists, it
        is returned. The branch structure is assumed as [-1, 0, 0, 1, 1, 2, 2, ...].

        Args:
            nbranches: Number of branches in the cell.
            ncomp: Number of compartments in each branch.
            copy: Whether to return a copy of the cell. Default is True.
            force_init: Force the init from scratch. Default is False.

        Returns:
            jx.Cell()."""
        if key := (nbranches, ncomp) not in cells or force_init:
            parents = [-1]
            depth = 0
            while nbranches > len(parents):
                parents = [-1] + [b // 2 for b in range(0, 2**depth - 2)]
                depth += 1
            parents = parents[:nbranches]
            branch = SimpleBranch(ncomp=ncomp, force_init=force_init)
            cells[key] = jx.Cell([branch] * nbranches, parents)
        return deepcopy(cells[key]) if copy and not force_init else cells[key]

    yield get_or_build_cell
    cells = {}


@pytest.fixture(scope="session")
def SimpleNet(SimpleCell):
    """Fixture for creating or retrieving an already created network."""
    nets = {}

    def get_or_build_net(
        ncells: int,
        nbranches: int,
        ncomp: int,
        connect: bool = False,
        copy: bool = True,
        force_init: bool = False,
    ) -> jx.Network:
        """Create or retrieve a network.

        If a network with the same number of cells, branches, compartments, and
        connections already exists, it is returned.

        Args:
            ncells: Number of cells in the network.
            nbranches: Number of branches in each cell.
            ncomp: Number of compartments in each branch.
            connect: Whether to connect the first two cells in the network.
            copy: Whether to return a copy of the network. Default is True.
            force_init: Force the init from scratch. Default is False.

        Returns:
            jx.Network()."""
        if key := (ncells, nbranches, ncomp, connect) not in nets or force_init:
            net = jx.Network(
                [SimpleCell(nbranches=nbranches, ncomp=ncomp, force_init=force_init)]
                * ncells
            )
            if connect:
                jx.connect(net[0, 0, 0], net[1, 0, 0], IonotropicSynapse())
            nets[key] = net
        return deepcopy(nets[key]) if copy and not force_init else nets[key]

    yield get_or_build_net
    nets = {}


@pytest.fixture(scope="session")
def SimpleMorphCell():
    """Fixture for creating or retrieving an already created morpholgy."""

    cells = {}

    def get_or_build_cell(
        fname: Optional[str] = None,
        ncomp: int = 1,
        max_branch_len: float = 2_000.0,
        ignore_swc_tracing_interruptions: bool = True,
        swc_backend: str = "graph",
        copy: bool = True,
        force_init: bool = False,
    ) -> jx.Cell:
        """Create or retrieve a cell from an SWC file.

        If a cell with the same SWC file, number of compartments, and maximum branch
        length already exists, it is returned.

        Args:
            fname: Path to the SWC file.
            ncomp: Number of compartments in each branch.
            max_branch_len: Maximum length of a branch.
            copy: Whether to return a copy of the cell. Default is True.
            force_init: Force the init from scratch. Default is False.

        Returns:
            jx.Cell()."""
        dirname = os.path.dirname(__file__)
        default_fname = os.path.join(dirname, "swc_files", "morph_ca1_n120.swc")
        fname = default_fname if fname is None else fname
        if key := (fname, ncomp, max_branch_len) not in cells or force_init:
            cells[key] = jx.read_swc(
                fname,
                ncomp=ncomp,
                max_branch_len=max_branch_len,
                assign_groups=True,
                backend=swc_backend,
                ignore_swc_tracing_interruptions=ignore_swc_tracing_interruptions,
            )
        return deepcopy(cells[key]) if copy and not force_init else cells[key]

    yield get_or_build_cell
    cells = {}


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


def pytest_collection_modifyitems(config, items):
    NEW_BASELINE = (
        int(os.environ["NEW_BASELINE"]) if "NEW_BASELINE" in os.environ else 0
    )

    dirname = os.path.dirname(__file__)
    baseline_fname = os.path.join(dirname, "regression_test_baselines.json")

    def should_skip_regression():
        return not NEW_BASELINE and not os.path.exists(baseline_fname)

    if should_skip_regression():
        for item in items:
            if "regression" in item.keywords:
                skip_regression = pytest.mark.skip(
                    reason="need NEW_BASELINE env to run"
                )
                item.add_marker(skip_regression)

    # Avoid that functions marked as `additional_neuron_tests` arent tested by default.
    #
    # If we run `pytest tests` then the `additional_neuron_tests` tests are not run.
    # This is useful because some tests based on NEURON (e.g.
    # `jaxley_vs_neuron/test_cell::test_similarity_ion_diffusion`) cause a segmentation
    # error downstream (in later tests). In addition, I have not set up a github action
    # to automatically compile NEURON channel models (which would also cause the tests
    # to fail. For these two reasons, we skip some NEURON tests by default and only run
    # them at release locally.
    #
    # Written by ChatGPT.
    if config.getoption("-m") == "additional_neuron_tests":
        return  # Run normally if explicitly requested

    skip_additional = pytest.mark.skip(
        reason="Skipped unless explicitly enabled with -m 'additional_neuron_tests'"
    )
    for item in items:
        if "additional_neuron_tests" in item.keywords:
            item.add_marker(skip_additional)


@pytest.fixture(scope="session", autouse=True)
def print_session_report(request, pytestconfig):
    """Cleanup a testing directory once we are finished."""
    NEW_BASELINE = (
        int(os.environ["NEW_BASELINE"]) if "NEW_BASELINE" in os.environ else 0
    )

    dirname = os.path.dirname(__file__)
    baseline_fname = os.path.join(dirname, "regression_test_baselines.json")
    results_fname = os.path.join(dirname, "regression_test_results.json")

    collected_regression_tests = [
        item for item in request.session.items if item.get_closest_marker("regression")
    ]

    def update_baseline():
        results = load_json(results_fname)
        with open(baseline_fname, "w") as f:
            json.dump(results, f, indent=2)
        os.remove(results_fname)

    def print_regression_report():
        baselines = load_json(baseline_fname)
        results = load_json(results_fname)

        report = generate_regression_report(baselines, results)
        # "No baselines found. Run `git checkout main;UPDATE_BASELINE=1 pytest -m regression; git checkout -`"
        with open(dirname + "/regression_test_report.txt", "w") as f:
            f.write(report)

        # the following allows to print the report to the console despite pytest
        # capturing the output and without specifying the "-s" flag
        capmanager = request.config.pluginmanager.getplugin("capturemanager")
        with capmanager.global_and_fixture_disabled():
            print("\n\n\nRegression Test Report\n----------------------\n")
            print(report)

    if len(collected_regression_tests) > 0:
        if NEW_BASELINE:
            request.addfinalizer(update_baseline)
        request.addfinalizer(print_regression_report)



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import networkx as nx
import numpy as np

from jaxley.io.graph import connect_graphs, from_graph
from jaxley.modules.base import to_graph


def morph_delete(module_view) -> "Cell":
    """Deletes part of a morphology.

    This function can only delete entire branches. It does not support deleting
    compartments of a branch.

    This function deletes all existing recordings, stimuli, and trainable parameters.

    Args:
        module_view: View of a `jx.Cell`. Defines the branches to be deleted.

    Returns:
        A cell in which specified branches are deleted.

    Example usage
    ^^^^^^^^^^^^^

    ::

        cell = jx.read_swc("path_to_swc_file.swc", ncomp=1)
        cell = morph_delete(cell.axon)

    ::

        cell = jx.Cell(branch, parents=[-1, 0, 0, 1, 1])
        cell = morph_delete(cell.branch([3, 4]))
    """
    # The `connect_graph` method cannot yet handle networks and branches.
    assert module_view.base.__class__.__name__ == "Cell", (
        f"You are trying to delete parts of a "
        "`jx.{module_view.base.__class__.__name__}`. "
        "Only `jx.Cell` is allowed in `morph_delete()`."
    )

    # Assert that there are no recordings, stimuli, and trainables.
    assert len(module_view.base.recordings) == 0, (
        f"Found {len(module_view.base.recordings)} recordings. This is not "
        f"supported. Please run `cell.delete_recordings()`."
    )
    assert len(module_view.base.externals) == 0, (
        f"Found {len(module_view.base.externals)} external states (stimuli or "
        f"clamps). This is not supported. Please run `cell.delete_stimuli()` or "
        f"`cell.delete_clamps()`."
    )
    assert len(module_view.base.trainable_params) == 0, (
        f"Found {len(module_view.base.trainable_params)} trainable parameters. "
        f"This is not supported. Please run `cell.delete_trainables()`."
    )

    # If the user did not run `compute_xyz` or `compute_compartment_centers`, we run
    # it automatically.
    if np.isnan(module_view.base.xyzr[0][0, 0]):
        module_view.base.compute_xyz()
    if "x" not in module_view.base.nodes.columns:
        module_view.base.compute_compartment_centers()

    comps_to_delete = module_view.nodes.index
    comp_graph = to_graph(module_view.base)

    nodes_to_keep = []
    for node in comp_graph.nodes:
        # We keep all `branchpoints`. The ones that have degree <2 will automatically
        # be trimmed away `_remove_branch_points_at_tips`, which is run during
        # `from_graph()`.
        if (
            comp_graph.nodes[node]["type"] == "branchpoint"
            or comp_graph.nodes[node]["comp_index"] not in comps_to_delete
        ):
            nodes_to_keep.append(node)

    comp_graph = nx.subgraph(comp_graph, nodes_to_keep)
    cell = from_graph(comp_graph)
    return cell


def morph_connect(module_view1, module_view2) -> "Cell":
    """Combines two morphologies into a single cell.

    Both morphologies must have the same number of compartments per branch in all
    branches.

    This function deletes all existing recordings, stimuli, and trainable parameters.

    Args:
        module_view1: View of a ``jx.Cell()``. Must have been created with a
            command ending on ``loc(0.0)`` or ``loc(1.0)``. For example, the following
            are valid:
            ``cell.branch(0).loc(0.0)``, ``cell.branch(5).loc(0.0)``,
            ``cell.branch(5).loc(1.0)``.
            But those are not valid:
            ``cell.branch(0).comp(0)`` (uses ``.comp``), ``cell.branch(5).loc(0.9)``
            (does not use ``loc(0.0)`` or ``loc(1.0)``).
        module_view2: The view of a ``jx.Cell()``. Must follow the same rules as
            ``module_view1``.

    Returns:
        A ``jx.Cell`` which is made up of both input cells.

    Example usage
    ^^^^^^^^^^^^^

    ::

        cell = jx.read_swc("path_to_swc_file.swc", ncomp=1)
        stub = jx.Cell()
        cell = morph_connect(cell.branch(0).loc(0.0), stub.branch(0).loc(0.0))
    """
    # The `connect_graph` method cannot yet handle networks and branches.
    for view in [module_view1, module_view2]:
        assert view.base.__class__.__name__ == "Cell", (
            f"You are trying to connect to a `jx.{view.base.__class__.__name__}`. "
            "Only `jx.Cell` is allowed in `morph_connect()`."
        )

    # Assert that there are no recordings, stimuli, and trainables.
    for view in [module_view1, module_view2]:
        assert len(view.base.recordings) == 0, (
            f"Found {len(view.base.recordings)} recordings. This is not "
            f"supported. Please run `cell.delete_recordings()`."
        )
        assert len(view.base.externals) == 0, (
            f"Found {len(view.base.externals)} external states (stimuli or "
            f"clamps). This is not supported. Please run `cell.delete_stimuli()` or "
            f"`cell.delete_clamps()`."
        )
        assert len(view.base.trainable_params) == 0, (
            f"Found {len(view.base.trainable_params)} trainable parameters. "
            f"This is not supported. Please run `cell.delete_trainables()`."
        )

    # If the user did not run `compute_xyz` or `compute_compartment_centers`, we run
    # it automatically.
    for view in [module_view1, module_view2]:
        if np.isnan(view.base.xyzr[0][0, 0]):
            view.base.compute_xyz()
        if "x" not in view.base.nodes.columns:
            view.base.compute_compartment_centers()

    graph1 = to_graph(module_view1.base, channels=True)
    graph2 = to_graph(module_view2.base, channels=True)

    comps = []
    for view in [module_view1, module_view2]:
        global_comp = int(view.nodes.index.to_numpy()[0])
        if len(view._comp_edges) == 0:
            # Connect to the tip of a branch.
            comps.append(global_comp)
        else:
            # Connect to a branchpoint.
            # Branchpoint counters are strictly higher than compartment counters,
            # so we can simply take the `max()` to get the branchpoint index.
            comps.append(int(view._comp_edges["sink"].max()))

    combined_graph = connect_graphs(graph1, graph2, comps[0], comps[1])
    cell = from_graph(combined_graph)
    return cell



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

from jaxley.connect import connect
from jaxley.synapses.ionotropic import IonotropicSynapse

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")
from typing import Optional

import jax.numpy as jnp
import numpy as np
import pytest

import jaxley as jx
from jaxley.channels import HH, CaL, CaT, Channel, K, Km, Leak, Na


def test_clamp_pointneuron(SimpleComp):
    comp = SimpleComp()
    comp.insert(HH())
    comp.record()
    comp.clamp("v", -50.0 * jnp.ones((1000,)))

    v = jx.integrate(comp, t_max=1.0)
    assert np.all(v[:, 1:] == -50.0)


def test_clamp_currents(SimpleComp):
    comp = SimpleComp()
    comp.insert(HH())
    comp.record("i_HH")

    # test clamp
    comp.clamp("i_HH", 1.0 * jnp.ones((1000,)))
    i1 = jx.integrate(comp, t_max=1.0)
    assert np.all(i1[:, 1:] == 1.0)

    # test data clamp
    data_clamps = None
    ipts = 1.0 * jnp.ones((1000,))
    data_clamps = comp.data_clamp("i_HH", ipts, data_clamps=data_clamps)

    i2 = jx.integrate(comp, data_clamps=data_clamps, t_max=1.0)
    assert np.all(i2[:, 1:] == 1.0)

    assert np.all(np.isclose(i1, i2))


def test_clamp_synapse(SimpleNet):
    net = SimpleNet(2, 1, 1)
    connect(net[0, 0, 0], net[1, 0, 0], IonotropicSynapse())
    net.record("IonotropicSynapse_s")

    # test clamp
    net.clamp("IonotropicSynapse_s", 1.0 * jnp.ones((1000,)))
    s1 = jx.integrate(net, t_max=1.0)
    assert np.all(s1[:, 1:] == 1.0)

    net.delete_clamps()

    # test data clamp
    data_clamps = None
    ipts = 1.0 * jnp.ones((1000,))
    data_clamps = net.data_clamp("IonotropicSynapse_s", ipts, data_clamps=data_clamps)

    s2 = jx.integrate(net, data_clamps=data_clamps, t_max=1.0)
    assert np.all(s2[:, 1:] == 1.0)

    assert np.all(np.isclose(s1, s2))


def test_clamp_multicompartment(SimpleBranch):
    branch = SimpleBranch(4)
    branch.insert(HH())
    branch.record()
    branch.comp(0).clamp("v", -50.0 * jnp.ones((1000,)))

    v = jx.integrate(branch, t_max=1.0)

    # The clamped compartment should be fixed.
    assert np.all(v[0, 1:] == -50.0)

    # For other compartments, the voltage should have non-zero std.
    assert np.all(np.std(v[1:, 1:], axis=1) > 0.1)


def test_clamp_and_stimulate_api(SimpleCell):
    """Ensure proper behaviour when `.clamp()` and `.stimulate()` are combined."""
    cell1 = SimpleCell(1, 4)
    cell2 = SimpleCell(1, 4)
    net = jx.Network([cell1, cell2])

    net.insert(HH())
    net[0, 0, 0].record()
    net[1, 0, 0].record()

    net[0, 0, 0].clamp("v", 0.1 * jnp.ones((1000,)))
    net[1, 0, 0].stimulate(0.1 * jnp.ones((1000,)))

    vs1 = jx.integrate(net, t_max=1.0)

    cell1.insert(HH())
    cell1[0, 0].record()
    cell1[0, 0].clamp("v", 0.1 * jnp.ones((1000,)))
    vs21 = jx.integrate(cell1, t_max=1.0)

    cell2.insert(HH())
    cell2[0, 0].record()
    cell2[0, 0].stimulate(0.1 * jnp.ones((1000,)))
    vs22 = jx.integrate(cell2, t_max=1.0)

    vs2 = jnp.concatenate([vs21, vs22])
    assert np.max(np.abs(vs1 - vs2)) < 1e-8


def test_data_clamp(SimpleComp):
    """Data clamp with no stimuli or data_stimuli, and no t_max (should get defined by the clamp)."""
    comp = SimpleComp()
    comp.insert(HH())
    comp.record()
    clamp = -50.0 * jnp.ones((1000,))

    def provide_data(clamp):
        data_clamps = comp.data_clamp("v", clamp)
        return data_clamps

    def simulate(clamp):
        data_clamps = provide_data(clamp)
        return jx.integrate(comp, data_clamps=data_clamps)

    jitted_simulate = jax.jit(simulate)

    s = jitted_simulate(clamp)
    assert np.all(s[:, 1:] == -50.0)


def test_data_clamp_and_data_stimulate(SimpleComp):
    """In theory people shouldn't use these two together, but at least it shouldn't break."""
    comp = SimpleComp()
    comp.insert(HH())
    comp.record()
    clamp = -50.0 * jnp.ones((1000,))
    stim = 0.1 * jnp.ones((1000,))

    def provide_data(clamp, stim):
        data_clamps = comp.data_clamp("v", clamp)
        data_stims = comp.data_stimulate(stim)
        return data_clamps, data_stims

    def simulate(clamp, stim):
        data_clamps, data_stims = provide_data(clamp, stim)
        return jx.integrate(comp, data_clamps=data_clamps, data_stimuli=data_stims)

    jitted_simulate = jax.jit(simulate)

    s = jitted_simulate(clamp, stim)
    assert np.all(s[:, 1:] == -50.0)


def test_data_clamp_and_stimulate(SimpleComp):
    """Test that data clamp overrides a previously set stimulus."""
    comp = SimpleComp()
    comp.insert(HH())
    comp.record()
    clamp = -50.0 * jnp.ones((1000,))
    stim = 0.1 * jnp.ones((800,))
    t_max = clamp.shape[0] * 0.025  # make sure the stimulus gets padded
    comp.stimulate(stim)

    def simulate(clamp):
        data_clamps = comp.data_clamp("v", clamp)  # should override the stimulation
        return jx.integrate(comp, data_clamps=data_clamps, t_max=t_max)

    jitted_simulate = jax.jit(simulate)

    s = jitted_simulate(clamp)
    assert np.all(s[:, 1:] == -50.0)


def test_data_clamp_and_clamp(SimpleComp):
    """Test that data clamp can override (same loc.) and add (another loc.) to clamp."""
    comp = SimpleComp()
    comp.insert(HH())
    comp.record()
    clamp1 = -50.0 * jnp.ones((1000,))
    clamp2 = -60.0 * jnp.ones((1000,))
    comp.clamp("v", clamp1)

    def simulate(clamp):
        data_clamps = comp.data_clamp(
            "v", clamp, None
        )  # should override the first clamp
        return jx.integrate(comp, data_clamps=data_clamps)

    jitted_simulate = jax.jit(simulate)

    # Clamp2 should override clamp1 here
    s = jitted_simulate(clamp2)
    assert np.all(s[:, 1:] == -60.0)

    comp2 = SimpleComp()
    comp2.insert(HH())
    branch1 = jx.Branch(comp, 4)
    branch2 = jx.Branch(comp2, 4)
    cell = jx.Cell([branch1, branch2], [-1, 0])

    # Apply the clamp1 to the second branch via clamp
    cell[1, 0].clamp("v", clamp1)

    cell.delete_recordings()
    cell.branch(0).comp(0).record()
    cell.branch(1).comp(0).record()

    def simulate(clamp):
        data_clamps = cell.branch(0).comp(0).data_clamp("v", clamp, None)
        return jx.integrate(cell, data_clamps=data_clamps)

    jitted_simulate = jax.jit(simulate)

    # Apply clamp2 to the first branch via data_clamp
    s = jitted_simulate(clamp2)

    assert np.all(s[0, 1:] == -60.0)
    assert np.all(s[1, 1:] == -50.0)



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")


import os

os.environ["XLA_PYTHON_CLIENT_MEM_FRACTION"] = ".4"

import jax.numpy as jnp
import numpy as np
import pytest
from jaxley_mech.channels.l5pc import CaHVA
from neuron import h, rxd

import jaxley as jx
from jaxley.channels import HH
from jaxley.pumps import CaFaradayConcentrationChange, CaNernstReversal

_ = h.load_file("stdlib.hoc")
_ = h.load_file("import3d.hoc")


@pytest.mark.parametrize("solver", ["bwd_euler", "crank_nicolson"])
def test_similarity(solver):
    """Test similarity of jaxley vs neuron."""
    i_delay = 3.0  # ms
    i_dur = 2.0  # ms
    i_amp = 1.0  # nA

    dt = 0.025  # ms
    t_max = 10.0  # ms

    voltages_jaxley = _run_jaxley(i_delay, i_dur, i_amp, dt, t_max, solver)
    voltages_neuron = _run_neuron(i_delay, i_dur, i_amp, dt, t_max, solver)

    assert np.mean(np.abs(voltages_jaxley - voltages_neuron)) < 0.05


def _run_jaxley(i_delay, i_dur, i_amp, dt, t_max, solver):
    ncomp_per_branch = 8
    comp = jx.Compartment()
    branch = jx.Branch(comp, ncomp_per_branch)
    cell = jx.Cell(branch, parents=[-1, 0, 0])
    cell.insert(HH())

    cell.set("radius", 5.0)
    cell.set("length", 10.0)
    cell.set("axial_resistivity", 1_000.0)
    cell.set("capacitance", 7.0)

    cell.set("HH_gNa", 0.120)
    cell.set("HH_gK", 0.036)
    cell.set("HH_gLeak", 0.0003)
    cell.set("HH_m", 0.07490098835688629)
    cell.set("HH_h", 0.488947681848153)
    cell.set("HH_n", 0.3644787002343737)
    cell.set("v", -62.0)

    current = jx.step_current(i_delay, i_dur, i_amp, dt, t_max)
    cell.branch(0).loc(0.0).stimulate(current)
    cell.branch(0).loc(0.0).record()
    cell.branch(1).loc(1.0).record()
    cell.branch(2).loc(1.0).record()

    voltages = jx.integrate(cell, delta_t=dt, solver=solver)
    return voltages


def _run_neuron(i_delay, i_dur, i_amp, dt, t_max, solver):
    if solver == "bwd_euler":
        h.secondorder = 0
    elif solver == "crank_nicolson":
        h.secondorder = 1
    else:
        raise ValueError

    ncomp_per_branch = 8
    h.dt = dt

    for sec in h.allsec():
        h.delete_section(sec=sec)

    branch1 = h.Section()
    branch2 = h.Section()
    branch3 = h.Section()

    branch2.connect(branch1, 1, 0)
    branch3.connect(branch1, 1, 0)

    for sec in h.allsec():
        sec.nseg = ncomp_per_branch

        sec.Ra = 1_000.0
        sec.L = 10.0 * ncomp_per_branch
        sec.diam = 2 * 5.0
        sec.cm = 7.0

        sec.insert("hh")
        sec.gnabar_hh = 0.120  # S/cm2
        sec.gkbar_hh = 0.036  # S/cm2
        sec.gl_hh = 0.0003  # S/cm2
        sec.ena = 50  # mV
        sec.ek = -77.0  # mV
        sec.el_hh = -54.3  # mV

    stim = h.IClamp(branch1(0.05))
    stim.delay = i_delay
    stim.dur = i_dur
    stim.amp = i_amp

    voltage1 = h.Vector()
    voltage1.record(branch1(0.05)._ref_v)
    voltage2 = h.Vector()
    voltage2.record(branch2(0.95)._ref_v)
    voltage3 = h.Vector()
    voltage3.record(branch3(0.95)._ref_v)

    v_init = -62.0

    def initialize():
        h.finitialize(v_init)
        h.fcurrent()

    def integrate():
        while h.t < t_max:
            h.fadvance()

    initialize()
    integrate()

    voltages = np.asarray([list(voltage1), list(voltage2), list(voltage3)])
    return voltages


def test_similarity_unequal_number_of_compartments():
    """Test similarity of jaxley vs neuron."""
    i_delay = 3.0  # ms
    i_dur = 2.0  # ms
    i_amp = 1.0  # nA

    dt = 0.025  # ms
    t_max = 10.0  # ms

    voltages_jaxley = _run_jaxley_unequal_ncomp(i_delay, i_dur, i_amp, dt, t_max)
    voltages_neuron = _run_neuron_unequal_ncomp(i_delay, i_dur, i_amp, dt, t_max)

    assert np.mean(np.abs(voltages_jaxley - voltages_neuron)) < 0.05


def _run_jaxley_unequal_ncomp(i_delay, i_dur, i_amp, dt, t_max):
    comp = jx.Compartment()
    branch1 = jx.Branch(comp, ncomp=1)
    branch2 = jx.Branch(comp, ncomp=2)
    branch3 = jx.Branch(comp, ncomp=3)
    branch4 = jx.Branch(comp, ncomp=4)
    cell = jx.Cell([branch1, branch2, branch3, branch4], parents=[-1, 0, 0, 1])
    cell.set("axial_resistivity", 10_000.0)
    cell.insert(HH())

    cell.set("radius", 5.0)
    cell.set("length", 20.0)
    cell.set("axial_resistivity", 1_000.0)
    cell.branch(1).set("capacitance", 10.0)
    cell.branch(3).set("capacitance", 20.0)

    cell.set("HH_gNa", 0.120)
    cell.set("HH_gK", 0.036)
    cell.set("HH_gLeak", 0.0003)
    cell.set("HH_m", 0.07490098835688629)
    cell.set("HH_h", 0.488947681848153)
    cell.set("HH_n", 0.3644787002343737)
    cell.set("v", -62.0)

    current = jx.step_current(i_delay, i_dur, i_amp, dt, t_max)
    cell.branch(1).comp(1).stimulate(current)
    cell.branch(0).comp(0).record()
    cell.branch(2).comp(2).record()
    cell.branch(3).comp(1).record()
    cell.branch(3).comp(3).record()

    voltages = jx.integrate(cell, delta_t=dt, voltage_solver="jaxley.dhs.cpu")
    return voltages


def _run_neuron_unequal_ncomp(i_delay, i_dur, i_amp, dt, t_max):
    h.secondorder = 0
    h.dt = dt

    for sec in h.allsec():
        h.delete_section(sec=sec)

    branch1 = h.Section()
    branch2 = h.Section()
    branch3 = h.Section()
    branch4 = h.Section()

    branch2.connect(branch1, 1, 0)
    branch3.connect(branch1, 1, 0)
    branch4.connect(branch2, 1, 0)

    ncomps = [1, 2, 3, 4]
    for i, sec in enumerate(h.allsec()):
        sec.nseg = ncomps[i]

        sec.Ra = 1_000.0
        sec.L = 20.0 * ncomps[i]
        sec.diam = 2 * 5.0

        sec.insert("hh")
        sec.gnabar_hh = 0.120  # S/cm2
        sec.gkbar_hh = 0.036  # S/cm2
        sec.gl_hh = 0.0003  # S/cm2
        sec.ena = 50  # mV
        sec.ek = -77.0  # mV
        sec.el_hh = -54.3  # mV

        if i == 1:
            sec.cm = 10.0
        if i == 3:
            sec.cm = 20.0

    stim = h.IClamp(branch2(0.6))  # The second out of two.
    stim.delay = i_delay
    stim.dur = i_dur
    stim.amp = i_amp

    voltage1 = h.Vector()
    voltage1.record(branch1(0.5)._ref_v)  # Only 1 compartment.
    voltage2 = h.Vector()
    voltage2.record(branch3(0.8)._ref_v)  # The third out of three compartments.
    voltage3 = h.Vector()
    voltage3.record(branch4(0.3)._ref_v)  # The second out of four comps.
    voltage4 = h.Vector()
    voltage4.record(branch4(0.9)._ref_v)  # The last out of four comps.

    v_init = -62.0

    def initialize():
        h.finitialize(v_init)
        h.fcurrent()

    def integrate():
        while h.t < t_max:
            h.fadvance()

    initialize()
    integrate()

    voltages = np.asarray(
        [list(voltage1), list(voltage2), list(voltage3), list(voltage4)]
    )
    return voltages


def _jaxley_ion_diffusion(
    dt, t_max, i_delay, i_dur, i_amp, v_init, diams, lengths, r_as
):
    comp = jx.Compartment()
    ncomp = 2
    branch = jx.Branch(comp, ncomp=ncomp)
    cell = jx.Cell(branch, parents=[-1, 0, 0])

    for i in range(3):
        cell.branch(i).set("radius", diams[i] / 2)
        cell.branch(i).set("axial_resistivity", r_as[i])
        cell.branch(i).set("length", lengths[i] / ncomp)

    cell.set("v", v_init)
    cell.branch(0).insert(HH())

    # Define calcium dynamics.
    cell.branch([0, 1]).insert(CaHVA())
    cell.insert(CaFaradayConcentrationChange())
    cell.insert(CaNernstReversal())
    cell.branch(0).set("CaHVA_gCaHVA", 0.00001)
    cell.branch(1).set("CaHVA_gCaHVA", 0.000005)
    cell.set("eCa", 0.0)

    cell.diffuse("CaCon_i")
    cell.set("axial_diffusion_CaCon_i", 1.0)

    cell.init_states()

    # Stimulate.
    step_current = jx.step_current(i_delay, i_dur, i_amp, dt, t_max)
    cell.branch(0).comp(0).stimulate(step_current)

    # Recordings of voltage and cacon_i at different compartments.
    cell.branch(0).comp(0).record("v")
    cell.branch(1).comp(1).record("v")
    cell.branch(2).comp(0).record("v")
    cell.branch(0).comp(1).record("CaCon_i")
    cell.branch(1).comp(0).record("CaCon_i")
    cell.branch(2).comp(1).record("CaCon_i")

    recordings = jx.integrate(cell, t_max=t_max, delta_t=dt)
    return recordings


def _neuron_ion_diffusion(
    dt, t_max, i_delay, i_dur, i_amp, v_init, diams, lengths, r_as
):
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "../arm64/.libs/libnrnmech.so")
    h.nrn_load_dll(fname)

    h.secondorder = 0
    h.dt = dt

    for sec in h.allsec():
        h.delete_section(sec=sec)

    branch0 = h.Section()
    branch1 = h.Section()
    branch2 = h.Section()

    branch0.insert("hh")

    for i, branch in enumerate([branch0, branch1, branch2]):
        branch.nseg = 2
        branch.Ra = r_as[i]
        branch.L = lengths[i]
        branch.diam = diams[i]

    branch0.insert("Ca_HVA")
    branch1.insert("Ca_HVA")
    branch0.gCa_HVAbar_Ca_HVA = 0.00001
    branch1.gCa_HVAbar_Ca_HVA = 0.000005

    branch1.connect(branch0, 1, 0)
    branch2.connect(branch0, 1, 0)

    # Set diffusion to zero. This is causing the different interpretation of `.area()`
    # of NEURON and therefore causes this test to fail (also for d>0) if the radiuses
    # of compartments are not the same (as is the case here).
    r = rxd.Region(h.allsec(), nrn_region="i", geometry=None)
    ca = rxd.Species(r, d=1, name="ca", charge=2)

    # Stimulate.
    stim = h.IClamp(branch0(0.3))
    stim.delay = i_delay
    stim.dur = i_dur
    stim.amp = i_amp

    # Voltage recordings.
    v0 = h.Vector()
    v0.record(branch0(0.3)._ref_v)
    v1 = h.Vector()
    v1.record(branch1(0.7)._ref_v)
    v2 = h.Vector()
    v2.record(branch2(0.3)._ref_v)

    # Calcium recordings.
    cacon0 = h.Vector()
    cacon0.record(branch0(0.7)._ref_cai)
    cacon1 = h.Vector()
    cacon1.record(branch1(0.3)._ref_cai)
    cacon2 = h.Vector()
    cacon2.record(branch2(0.7)._ref_cai)

    def initialize():
        h.finitialize(v_init)
        h.fcurrent()

    def integrate():
        while h.t < t_max:
            h.fadvance()

    initialize()
    integrate()

    v0 = np.asarray(list(v0))
    v1 = np.asarray(list(v1))
    v2 = np.asarray(list(v2))
    cacon0 = np.asarray(list(cacon0))
    cacon1 = np.asarray(list(cacon1))
    cacon2 = np.asarray(list(cacon2))

    return np.stack([v0, v1, v2, cacon0, cacon1, cacon2])


def test_similarity_ion_diffusion():
    """Test whether ion diffusion behaves as in NEURON.

    This test works, but it can cause a segmentation fault on my MacBook (locally).
    The segmentation fault does not happen while running this, but it happens in
    one of the later calls to NEURON (e.g. in `test_comp.py`). In addition, this test
    relies on having compiled NEURON channel models, which I have not yet set up a
    Github action for.

    This test is closely related to the test
    `pytest test_branch.py -k test_similarity_voltage_clamp_cacon_i_behavior`
    which is skipped (because it fails for different diams between compartments.)
    """
    t_max = 1_000.01
    dt = 0.025

    # The test passes only when diams has the same values. See also the skipped test in
    # `pytest test_branch.py -k test_similarity_voltage_clamp_cacon_i_behavior`
    diams = [8.5, 8.5, 8.5]
    lengths = [100.0, 50.0, 70.0]  # Lengths of each branch.
    r_as = [1_000.0, 500.0, 750.0]  # Axial voltage resistivity of each branch.
    v_init = -65.0
    i_delay = 50.0
    i_dur = 20.0
    i_amp = 0.1

    v_and_cacon_jaxley = _jaxley_ion_diffusion(
        dt, t_max, i_delay, i_dur, i_amp, v_init, diams, lengths, r_as
    )
    # We are not running NEURON, and instead just saving its voltages below. This
    # is because rxd sometimes caused a segmentation fault in our tests.
    # v_and_cacon_neuron = _neuron_ion_diffusion(
    #     dt, t_max, i_delay, i_dur, i_amp, v_init, diams, lengths, r_as
    # )

    # Analysis.
    v_jaxley = v_and_cacon_jaxley[:3]
    cacon_jaxley = v_and_cacon_jaxley[3:]
    # v_neuron = v_and_cacon_neuron[:3]
    # cacon_neuron = v_and_cacon_neuron[3:]

    v_neuron = jnp.asarray(
        [
            [
                -64.9740566,
                -60.88748118,
                -57.9800073,
                -54.45089083,
                21.87809771,
                -73.1579158,
                -73.58213181,
                -70.61212201,
                -67.2714086,
                -67.58651223,
            ],
            [
                -64.97405911,
                -61.7164889,
                -58.78402151,
                -55.49142503,
                13.19733503,
                -69.01489137,
                -74.21113676,
                -71.42037101,
                -68.0896641,
                -67.67389774,
            ],
            [
                -64.97405923,
                -61.73392024,
                -58.79977817,
                -55.51516832,
                12.12792837,
                -68.73055718,
                -74.22165672,
                -71.43820274,
                -68.10784275,
                -67.67786202,
            ],
        ]
    )
    cacon_neuron_spike = (
        jnp.asarray(
            [
                [
                    5.00000453,
                    5.00000486,
                    5.00000626,
                    5.0000127,
                    5.1048348,
                    8.79260654,
                    9.30190049,
                    9.30284364,
                    9.29821517,
                    9.29349743,
                ],
                [
                    5.00000227,
                    5.00000243,
                    5.00000307,
                    5.00000599,
                    5.02163309,
                    6.81641896,
                    7.16615776,
                    7.16847423,
                    7.16688202,
                    7.16524459,
                ],
                [
                    5.0,
                    5.0,
                    5.0,
                    5.0,
                    5.0,
                    5.00000213,
                    5.00001658,
                    5.00004703,
                    5.00009347,
                    5.00015577,
                ],
            ]
        )
        * 1e-5
    )
    cacon_neuron_end = (
        jnp.asarray(
            [
                [
                    8.30158656,
                    8.29842432,
                    8.29527959,
                    8.29215224,
                    8.28904216,
                    8.28594922,
                    8.2828733,
                    8.27981429,
                    8.27677206,
                    8.2737465,
                ],
                [
                    7.10557499,
                    7.10586209,
                    7.10614861,
                    7.10643455,
                    7.10671989,
                    7.10700463,
                    7.10728874,
                    7.10757223,
                    7.10785509,
                    7.10813731,
                ],
                [
                    5.46553328,
                    5.4689818,
                    5.47242902,
                    5.47587487,
                    5.47931927,
                    5.48276214,
                    5.48620342,
                    5.48964301,
                    5.49308086,
                    5.49651688,
                ],
            ]
        )
        * 1e-5
    )

    v_max_error = np.max(np.abs(v_jaxley[:, 2_000:3_000:100] - v_neuron))
    cacon_max_error_spike = np.max(
        np.abs(cacon_jaxley[:, 2_000:3_000:100] - cacon_neuron_spike)
    )
    cacon_max_error_end = np.max(np.abs(cacon_jaxley[:, -2000::200] - cacon_neuron_end))
    # Note: for very thin cables (<1um radius), this test can fail by a tiny bit, even
    # if both comps have the same diameter.
    assert cacon_max_error_spike < 2e-6, f"CaCon_i error {cacon_max_error_spike}"
    assert cacon_max_error_end < 2e-6, f"CaCon_i error {cacon_max_error_end}"

    # To be on the safe side, we also check voltages. However, this test is generally
    # about CaCon_i.
    assert v_max_error < 1e-2, f"Voltage error {v_max_error}"



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import Optional
from warnings import warn

import jax
import jax.numpy as jnp

from jaxley.channels import Channel


class Fire(Channel):
    """Mechanism to reset the voltage when it crosses a threshold.

    When combined with a ``Leak`` channel, this can be used to implement
    leaky-integrate-and-fire neurons.

    Note that, after the voltage is reset by this channel, other channels (or external
    currents), can still modify the membrane voltage `within the same time step`.
    """

    def __init__(self, name: Optional[str] = None):
        self.current_is_in_mA_per_cm2 = True
        super().__init__(name)
        self.channel_params = {f"{self.name}_vth": -50, f"{self.name}_vreset": -70}
        self.channel_states = {f"{self.name}_spikes": False}
        self.current_name = f"{self.name}_fire"
        warn(
            "The `Fire` channel does not support surrogate gradients. Its gradient "
            "will be zero after every spike."
        )

    def update_states(self, states, dt, v, params):
        """Reset the voltage when a spike occurs and log the spike"""
        prefix = self._name
        vreset = params[f"{prefix}_vreset"]
        vth = params[f"{prefix}_vth"]

        spike_occurred = v > vth
        v = jax.lax.select(spike_occurred, vreset, v)
        return {"v": v, f"{self.name}_spikes": spike_occurred}

    def compute_current(self, states, v, params):
        return jnp.zeros((1,))

    def init_state(self, states, v, params, delta_t):
        return {}



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>
from collections.abc import Callable
from math import prod
from typing import Callable, Dict, List, Optional, Tuple, Union

import jax
import jax.numpy as jnp
import pandas as pd
from jax import Array
from jax.typing import ArrayLike

from jaxley.modules import Module
from jaxley.utils.cell_utils import params_to_pstate
from jaxley.utils.jax_utils import nested_checkpoint_scan


def build_init_and_step_fn(
    module: Module,
    voltage_solver: str = "jaxley.dhs",
    solver: str = "bwd_euler",
) -> Tuple[Callable, Callable]:
    """Return ``init_fn`` and ``step_fn`` which initialize modules and run update steps.

    This method can be used to gain additional control over the simulation workflow.
    It exposes the ``step`` function, which can be used to perform step-by-step updates
    of the differential equations.

    Args:
        module: A `Module` object that e.g. a cell.
        voltage_solver: Voltage solver used in step. Defaults to "jaxley.stone".
        solver: ODE solver. One of
            { "bwd_euler" | "crank_nicolson" "| "fwd_euler" | "exp_euler" }.

    Returns:
        init_fn, step_fn: Functions that initialize the state and parameters, and
            perform a single integration step, respectively.

    Example usage
    ^^^^^^^^^^^^^

    The following allows you to perform a step-by-step update of the differential
    equations.

    ::

        import jax.numpy as jnp
        import jaxley as jx
        from jaxley.integrate import build_init_and_step_fn

        t_max = 3.0
        delta_t = 0.025

        cell = jx.Cell()
        cell.record()
        cell.stimulate(jx.step_current(0, 1, 2, delta_t, t_max))
        params = cell.get_parameters()

        cell.to_jax()
        rec_inds = cell.recordings.rec_index.to_numpy()
        rec_states = cell.recordings.state.to_numpy()
        externals = cell.externals.copy()
        external_inds = cell.external_inds.copy()

        # Uncomment this line if `data_stimuli` is not `None`.
        # externals, external_inds = add_stimuli(externals, external_inds, data_stimuli)

        # Uncomment this line if `data_clamps` is not `None`.
        # externals, external_inds = add_clamps(externals, external_inds, data_clamps)

        # Initialize.
        init_fn, step_fn = build_init_and_step_fn(cell)
        states, params = init_fn(params)
        recordings = [
            states[rec_state][rec_ind][None]
            for rec_state, rec_ind in zip(rec_states, rec_inds)
        ]

        # Loop over the ODE. The `step_fn` can be jitted for improving speed.
        steps = int(t_max / delta_t)  # Steps to integrate
        for step in range(steps):
            # Get externals at current timestep.
            externals_now = {}
            for key in externals.keys():
                externals_now[key] = externals[key][:, step]

            states = step_fn(
                states, params, externals_now, external_inds, delta_t=delta_t
            )
            recs = jnp.asarray(
                [
                    states[rec_state][rec_ind]
                    for rec_state, rec_ind in zip(rec_states, rec_inds)
                ]
            )
            recordings.append(recs)

        rec = jnp.stack(recordings, axis=0).T

    """
    # Initialize the external inputs and their indices.
    external_inds = module.external_inds.copy()

    def init_fn(
        params: list[dict[str, Array]],
        all_states: dict | None = None,
        param_state: list[dict] | None = None,
        delta_t: float = 0.025,
    ) -> Tuple[Dict, Dict]:
        """Initializes the parameters and states of the neuron model.

        Args:
            params: List of trainable parameters.
            all_states: State if already initialized. Defaults to None.
            param_state: Parameters returned by `data_set`.. Defaults to None.
            delta_t: Step size. Defaults to 0.025.

        Returns:
            All states and parameters.
        """
        # Make the `trainable_params` of the same shape as the `param_state`, such that
        # they can be processed together by `get_all_parameters`.
        pstate = params_to_pstate(params, module.indices_set_by_trainables)
        if param_state is not None:
            pstate += param_state

        all_params = module.get_all_parameters(pstate)
        if all_states is None:
            all_states = module.get_all_states(pstate)
            all_states = module.append_channel_currents_to_states(
                all_states, all_params, delta_t
            )
        return all_states, all_params

    def step_fn(
        all_states: Dict,
        all_params: Dict,
        externals: Dict,
        external_inds: Dict = external_inds,
        delta_t: float = 0.025,
    ) -> Dict:
        """Performs a single integration step with step size delta_t.

        Args:
            all_states: Current state of the neuron model.
            all_params: Current parameters of the neuron model.
            externals: External inputs.
            external_inds: External indices. Defaults to `module.external_inds`.
            delta_t: Time step. Defaults to 0.025.

        Returns:
            Updated states.
        """
        state = all_states
        state = module.step(
            state,
            delta_t,
            external_inds,
            externals,
            params=all_params,
            solver=solver,
            voltage_solver=voltage_solver,
        )
        return state

    return init_fn, step_fn


def add_stimuli(
    externals: dict,
    external_inds: dict,
    data_stimuli: tuple[ArrayLike, pd.DataFrame] | None = None,
) -> tuple[dict, dict]:
    """Extends the external inputs with the stimuli.

    Args:
        externals: Current external inputs.
        external_inds: Current external indices.
        data_stimuli: Additional data stimuli. Defaults to None.

    Returns:
        Updated external inputs and indices.
    """
    # If stimulus is inserted, add it to the external inputs.
    if "i" in externals.keys() or data_stimuli is not None:
        if "i" in externals.keys():
            if data_stimuli is not None:
                externals["i"] = jnp.concatenate([externals["i"], data_stimuli[1]])
                external_inds["i"] = jnp.concatenate(
                    [external_inds["i"], data_stimuli[2].index.to_numpy()]
                )
        else:
            externals["i"] = data_stimuli[1]
            external_inds["i"] = data_stimuli[2].index.to_numpy()

    return externals, external_inds


def add_clamps(
    externals: dict,
    external_inds: dict,
    data_clamps: tuple[str, ArrayLike, pd.DataFrame] | None = None,
) -> tuple[dict, dict]:
    """Adds clamps to the external inputs.

    Args:
        externals: Current external inputs.
        external_inds: Current external indices.
        data_clamps: Additional data clamps. Defaults to None.

    Returns:
        Updated external inputs and indices.
    """
    # If a clamp is inserted, add it to the external inputs.
    if data_clamps is not None:
        state_name, clamps, inds = data_clamps
        if state_name in externals.keys():
            externals[state_name] = jnp.concatenate([externals[state_name], clamps])
            external_inds[state_name] = jnp.concatenate(
                [external_inds[state_name], inds.index.to_numpy()]
            )
        else:
            externals[state_name] = clamps
            external_inds[state_name] = inds.index.to_numpy()

    return externals, external_inds


def integrate(
    module: Module,
    params: list[dict[str, Array]] = [],
    *,
    param_state: list[dict] | None = None,
    data_stimuli: tuple[ArrayLike, pd.DataFrame] | None = None,
    data_clamps: tuple[str, ArrayLike, pd.DataFrame] | None = None,
    t_max: float | None = None,
    delta_t: float = 0.025,
    solver: str = "bwd_euler",
    voltage_solver: str = "jaxley.dhs",
    checkpoint_lengths: Optional[List[int]] = None,
    all_states: Optional[Dict] = None,
    return_states: bool = False,
) -> Array:
    """Solves ODE and simulates neuron model.

    Args:
        params: Trainable parameters returned by `get_parameters()`.
        param_state: Parameters returned by `data_set`.
        data_stimuli: Outputs of `.data_stimulate()`, only needed if stimuli change
            across function calls.
        data_clamps: Outputs of `.data_clamp()`, only needed if clamps change across
            function calls.
        t_max: Duration of the simulation in milliseconds. If `t_max` is greater than
            the length of the stimulus input, the stimulus will be padded at the end
            with zeros. If `t_max` is smaller, then the stimulus with be truncated.
        delta_t: Time step of the solver in milliseconds.
        solver: Which ODE solver to use. Either of ["fwd_euler", "bwd_euler",
            "crank_nicolson", "exp_euler"]. Note that `exp_euler` is still
            experimental and is not yet optimized for networks.
        voltage_solver: Algorithm to solve quasi-tridiagonal linear system describing
            the voltage equations. The different options only take effect when
            `solver` is either `bwd_euler` or `crank_nicolson`. The options for
            `voltage_solver` are `jaxley.dhs` and `jax.sparse`.
            For unbranched cables, we also support `jaxley.stone` (which has good
            performance for unbranched cables on GPU).
        checkpoint_lengths: Number of timesteps at every level of checkpointing. The
            `prod(checkpoint_lengths)` must be larger or equal to the desired number of
            simulated timesteps. Warning: the simulation is run for
            `prod(checkpoint_lengths)` timesteps, and the result is posthoc truncated
            to the desired simulation length. Therefore, a poor choice of
            `checkpoint_lengths` can lead to longer simulation time. If `None`, no
            checkpointing is applied.
        all_states: An optional initial state that was returned by a previous
            `jx.integrate(..., return_states=True)` run. Overrides potentially
            trainable initial states.
        return_states: If True, it returns all states such that the current state of
            the `Module` can be set with `set_states`.

    Example usage
    ^^^^^^^^^^^^^

    The most simple usage is the following:

    ::

        cell = jx.Cell()
        v = jx.integrate(cell, t_max=10.0)

    If ``t_max`` is not passed, then you must have inserted a stimulus, and ``t_max``
    will match the stimulus length.

    Customizing the solver
    ^^^^^^^^^^^^^^^^^^^^^^

    If you use ``jx.integrate(..., voltage_solver="jaxley.dhs")``, we automatically
    choose between a CPU and a GPU optimized version. If you manually want to run the
    CPU-optimized version on GPU, do:

    ::

        cell._init_morph_jaxley_dhs_solve(allowed_nodes_per_level=1)
        v = jx.integrate(cell, voltage_solver="jaxley.dhs.cpu")

    To run the GPU-opotimized version on CPU, do:

    ::

        cell._init_morph_jaxley_dhs_solve(allowed_nodes_per_level=16)
        v = jx.integrate(cell, voltage_solver="jaxley.dhs.gpu")

    """
    if voltage_solver == "jaxley.dhs":
        # Automatically infer the voltage solver.
        if module._solver_device in ["gpu", "tpu"]:
            voltage_solver = "jaxley.dhs.gpu"
        else:
            voltage_solver = "jaxley.dhs.cpu"

    assert module.initialized, "Module is not initialized, run `._initialize()`."
    module.to_jax()  # Creates `.jaxnodes` from `.nodes` and `.jaxedges` from `.edges`.

    # Initialize the external inputs and their indices.
    externals = module.externals.copy()
    external_inds = module.external_inds.copy()

    # If stimulus is inserted, add it to the external inputs.
    externals, external_inds = add_stimuli(externals, external_inds, data_stimuli)

    # If a clamp is inserted, add it to the external inputs.
    externals, external_inds = add_clamps(externals, external_inds, data_clamps)

    if not externals.keys():
        # No stimulus was inserted and no clamp was set.
        assert t_max is not None, (
            "If no stimulus or clamp are inserted you have to specify the simulation"
            "duration at `jx.integrate(..., t_max=)`."
        )

    for key in externals.keys():
        externals[key] = externals[key].T  # Shape `(time, num_stimuli)`.

    if module.recordings.empty:
        raise ValueError("No recordings are set. Please set them.")
    rec_inds = module.recordings.rec_index.to_numpy()
    rec_states = module.recordings.state.to_numpy()

    # Shorten or pad stimulus depending on `t_max`.
    if t_max is not None:
        t_max_steps = int(t_max // delta_t + 1)

        # Pad or truncate the stimulus.
        for key in externals.keys():
            if t_max_steps > externals[key].shape[0]:
                if key == "i":
                    pad = jnp.zeros(
                        (t_max_steps - externals["i"].shape[0], externals["i"].shape[1])
                    )
                    externals["i"] = jnp.concatenate((externals["i"], pad))
                else:
                    raise NotImplementedError(
                        "clamp must be at least as long as simulation."
                    )
            else:
                externals[key] = externals[key][:t_max_steps, :]

    init_fn, step_fn = build_init_and_step_fn(
        module, voltage_solver=voltage_solver, solver=solver
    )
    all_states, all_params = init_fn(params, all_states, param_state, delta_t)

    if solver == "exp_euler":
        exp_matrix = module.solver_customizers["exp_euler"]["exp_euler_transition"]
        if exp_matrix is None:
            exp_matrix = module.build_exp_euler_transition_matrix(
                delta_t=delta_t,
                axial_conductances=all_params["axial_conductances"],
            )
        all_params["exp_euler_transition"] = exp_matrix

    def _body_fun(state, externals):
        state = step_fn(state, all_params, externals, external_inds, delta_t)
        recs = jnp.asarray(
            [
                state[rec_state][rec_ind]
                for rec_state, rec_ind in zip(rec_states, rec_inds)
            ]
        )
        return state, recs

    # If necessary, pad the stimulus with zeros in order to simulate sufficiently long.
    # The total simulation length will be `prod(checkpoint_lengths)`. At the end, we
    # return only the first `nsteps_to_return` elements (plus the initial state).
    if externals:
        example_key = list(externals.keys())[0]
        nsteps_to_return = len(externals[example_key])
    else:
        nsteps_to_return = t_max_steps

    if checkpoint_lengths is None:
        checkpoint_lengths = [nsteps_to_return]
        length = nsteps_to_return
    else:
        length = prod(checkpoint_lengths)
        size_difference = length - nsteps_to_return
        assert (
            nsteps_to_return <= length
        ), "The desired simulation duration is longer than `prod(nested_length)`."
        if externals:
            dummy_external = jnp.zeros(
                (size_difference, externals[example_key].shape[1])
            )
            for key in externals.keys():
                externals[key] = jnp.concatenate([externals[key], dummy_external])

    # Record the initial state.
    init_recs = jnp.asarray(
        [
            all_states[rec_state][rec_ind]
            for rec_state, rec_ind in zip(rec_states, rec_inds)
        ]
    )
    init_recording = jnp.expand_dims(init_recs, axis=0)

    # Run simulation.
    all_states, recordings = nested_checkpoint_scan(
        _body_fun,
        all_states,
        externals,
        length=length,
        nested_lengths=checkpoint_lengths,
    )
    recs = jnp.concatenate([init_recording, recordings[:nsteps_to_return]], axis=0).T
    return (recs, all_states) if return_states else recs



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>
from typing import Optional

import jax
from jax import Array
from jax.typing import ArrayLike

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

import jax.numpy as jnp
import numpy as np

import jaxley as jx
from jaxley.channels import Channel
from jaxley.solver_gate import exponential_euler, save_exp, solve_gate_exponential


class Dummy1(Channel):
    """A dummy channel which simply accumulates a state (same state as dummy2)."""

    def __init__(self, name: Optional[str] = None):
        self.current_is_in_mA_per_cm2 = True
        super().__init__(name)
        self.channel_params = {}
        self.channel_states = {"Dummy_s": 0.0}
        self.current_name = f"i_Dummy1"

    @staticmethod
    def update_states(u, dt, voltages, params):
        """Update state."""
        dummy_state = u["Dummy_s"]
        print("dummy_state1", dummy_state)
        return {"Dummy_s": dummy_state + 0.01}

    @staticmethod
    def compute_current(u, voltages, params):
        """Return current."""
        return jnp.zeros_like(voltages)


class Dummy2(Channel):
    """A dummy channel which simply accumulates a state (same state as dummy1)."""

    def __init__(self, name: Optional[str] = None):
        self.current_is_in_mA_per_cm2 = True
        super().__init__(name)
        self.channel_params = {}
        self.channel_states = {"Dummy_s": 0.0}
        self.current_name = f"i_Dummy2"

    @staticmethod
    def update_states(u, dt, voltages, params):
        """Update state."""
        dummy_state = u["Dummy_s"]
        print("dummy_state2", dummy_state)
        return {"Dummy_s": dummy_state + 0.01}

    @staticmethod
    def compute_current(u, voltages, params):
        """Return current."""
        return jnp.zeros_like(voltages)


class CaHVA(Channel):
    """High-Voltage-Activated (HVA) Ca2+ channel"""

    def __init__(self, name: Optional[str] = None):
        self.current_is_in_mA_per_cm2 = True
        super().__init__(name)
        self.channel_params = {
            f"{self._name}_gCaHVA": 0.00001,  # S/cm^2
        }
        self.channel_states = {
            f"{self._name}_m": 0.1,  # Initial value for m gating variable
            f"{self._name}_h": 0.1,  # Initial value for h gating variable
            "eCa": 0.0,  # mV, assuming eca for demonstration
        }
        self.current_name = f"i_Ca"
        self.META = {
            "reference": "Reuveni, Friedman, Amitai, and Gutnick, J.Neurosci. 1993",
            "mechanism": "HVA Ca2+ channel",
        }

    def update_states(
        self,
        u: dict[str, ArrayLike],
        dt: float,
        voltages: float,
        params: dict[str, Array],
    ):
        """Update state of gating variables."""
        prefix = self._name
        ms, hs = u[f"{prefix}_m"], u[f"{prefix}_h"]
        m_new = solve_gate_exponential(ms, dt, *self.m_gate(voltages))
        h_new = solve_gate_exponential(hs, dt, *self.h_gate(voltages))
        return {f"{prefix}_m": m_new, f"{prefix}_h": h_new, "eCa": u["eCa"]}

    def compute_current(
        self, u: dict[str, ArrayLike], voltages, params: dict[str, ArrayLike]
    ):
        """Compute the current through the channel."""
        prefix = self._name
        ms, hs = u[f"{prefix}_m"], u[f"{prefix}_h"]
        ca_cond = params[f"{prefix}_gCaHVA"] * (ms**2) * hs * 1000
        current = ca_cond * (voltages - u["eCa"])
        return current

    def init_state(self, voltages, params):
        """Initialize the state such at fixed point of gate dynamics."""
        prefix = self._name
        alpha_m, beta_m = self.m_gate(voltages)
        alpha_h, beta_h = self.h_gate(voltages)
        return {
            f"{prefix}_m": alpha_m / (alpha_m + beta_m),
            f"{prefix}_h": alpha_h / (alpha_h + beta_h),
        }

    @staticmethod
    def m_gate(v):
        """Voltage-dependent dynamics for the m gating variable."""
        alpha = (0.055 * (-27 - v + 1e-6)) / (save_exp((-27.0 - v + 1e-6) / 3.8) - 1.0)
        beta = 0.94 * save_exp((-75.0 - v + 1e-6) / 17.0)
        return alpha, beta

    @staticmethod
    def h_gate(v):
        """Voltage-dependent dynamics for the h gating variable."""
        alpha = 0.000457 * save_exp((-13.0 - v) / 50.0)
        beta = 0.0065 / (save_exp((-v - 15.0) / 28.0) + 1.0)
        return alpha, beta


class CaPump(Channel):
    """Calcium dynamics tracking inside calcium concentration, modeled after Destexhe et al. 1994."""

    def __init__(
        self,
        name: Optional[str] = None,
    ):
        self.current_is_in_mA_per_cm2 = True
        super().__init__(name)
        self.channel_params = {
            f"{self._name}_gamma": 0.05,  # Fraction of free calcium (not buffered)
            f"{self._name}_decay": 80,  # Rate of removal of calcium in ms
            f"{self._name}_depth": 0.1,  # Depth of shell in um
            f"{self._name}_minCai": 1e-4,  # Minimum intracellular calcium concentration in mM
        }
        self.channel_states = {
            f"CaCon_i": 5e-05,  # Initial internal calcium concentration in mM
        }
        self.current_name = f"i_Ca"
        self.META = {
            "reference": "Modified from Destexhe et al., 1994",
            "mechanism": "Calcium dynamics",
        }

    def update_states(self, u, dt, voltages, params):
        """Update internal calcium concentration based on calcium current and decay."""
        prefix = self._name
        ica = u["i_Ca"] / 1_000.0
        cai = u["CaCon_i"]
        gamma = params[f"{prefix}_gamma"]
        decay = params[f"{prefix}_decay"]
        depth = params[f"{prefix}_depth"]
        minCai = params[f"{prefix}_minCai"]

        FARADAY = 96485  # Coulombs per mole

        # Calculate the contribution of calcium currents to cai change
        drive_channel = -10_000.0 * ica * gamma / (2 * FARADAY * depth)

        cai_tau = decay
        cai_inf = minCai + decay * drive_channel
        new_cai = exponential_euler(cai, dt, cai_inf, cai_tau)

        return {f"CaCon_i": new_cai}

    def compute_current(self, u, voltages, params):
        """This dynamics model does not directly contribute to the membrane current."""
        return 0

    def init_state(self, voltages, params):
        """Initialize the state at fixed point of gate dynamics."""
        return {}


def test_shared_state():
    """Test whether two channels can share a state.

    This has to be copied into a notebook and executed with `jax.disable_jit():`."""
    comp1 = jx.Compartment()
    comp1.insert(Dummy1())

    comp2 = jx.Compartment()
    comp2.insert(Dummy2())

    comp3 = jx.Compartment()
    comp3.insert(Dummy1())
    comp3.insert(Dummy2())

    voltages = []
    for comp in [comp1, comp2, comp3]:
        comp.record()
        current = jx.step_current(
            i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
        )
        comp.stimulate(current)

        voltages.append(jx.integrate(comp))


def test_current_as_state_multicompartment():
    """#323 had discovered a bug when currents are only used in a few compartments."""
    comp = jx.Compartment()
    branch = jx.Branch(comp, 2)

    branch.comp(0).insert(CaHVA())  # defines current `i_Ca`
    branch.comp(0).insert(CaPump())  # uses `states["i_Ca"]`

    branch.comp(0).record()
    _ = jx.integrate(branch, t_max=1.0)



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

from typing import Optional

import jax.numpy as jnp

from jaxley.channels import Channel
from jaxley.solver_gate import exponential_euler


class Rate(Channel):
    """Rate-based, unit-less, neuron model."""

    def __init__(self, name: Optional[str] = None):
        self.current_is_in_mA_per_cm2 = True
        super().__init__(name)
        self.channel_params = {f"{self.name}_tau": 1.0}
        self.channel_states = {}
        self.current_name = f"{self.name}_rate"

    def update_states(self, states, dt, v, params):
        """Voltages get pulled towards zero."""
        tau = params[f"{self.name}_tau"]
        return {"v": exponential_euler(v, dt, 0.0, tau)}

    def compute_current(self, states, v, params):
        return jnp.zeros((1,))

    def init_state(self, states, v, params, delta_t):
        return {}



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

import os

os.environ["XLA_PYTHON_CLIENT_MEM_FRACTION"] = ".8"
import numpy as np
import pytest
from neuron import h

import jaxley as jx

_ = h.load_file("stdlib.hoc")
_ = h.load_file("import3d.hoc")


# Test is failing for "morph_ca1_n120.swc". This is because NEURON and Jaxley handle
# interrupted soma differently, see issue #140.
@pytest.mark.parametrize(
    "file", ["morph_ca1_n120_single_point_soma.swc", "morph_minimal.swc"]
)
def test_swc_reader_lengths(file):
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", file)

    cell = jx.read_swc(fname, ncomp=1)
    pathlengths = cell.nodes.length.to_numpy()
    if pathlengths[0] == 0.1:
        pathlengths = pathlengths[1:]

    for sec in h.allsec():
        h.delete_section(sec=sec)

    cell = h.Import3d_SWC_read()
    cell.input(fname)
    i3d = h.Import3d_GUI(cell, False)
    i3d.instantiate(None)

    neuron_pathlengths = []
    for sec in h.allsec():
        neuron_pathlengths.append(sec.L)
    neuron_pathlengths = np.asarray(neuron_pathlengths)

    for p in pathlengths:
        dists = np.abs(neuron_pathlengths - p)
        assert np.min(dists) < 1e-3, "Some branches have too large distance."

    assert len(pathlengths) == len(
        neuron_pathlengths
    ), "Number of branches does not match."


def test_dummy_compartment_length():
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", "morph_soma_both_ends.swc")

    cell = jx.read_swc(fname, ncomp=1)
    pathlengths = cell.nodes.length.to_numpy().tolist()
    parents = cell.comb_parents.tolist()
    assert parents == [-1, 0, 1]
    assert pathlengths == [2.2, 1.0, 2.6]


@pytest.mark.parametrize(
    "file", ["morph_ca1_n120_250_single_point_soma.swc", "morph_ca1_n120_250.swc"]
)
def test_swc_radius(file):
    """We expect them to match for sufficiently large ncomp. See #140."""
    ncomp = 64

    # Can not use full morphology because of branch sorting.
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", file)

    # _, _, radius_fns, _, _ = swc2jaxley(fname, max_branch_len=2000.0, sort=False)
    cell = jx.read_swc(fname, ncomp=ncomp)
    jaxley_diams = cell.nodes.radius.to_numpy() * 2
    jaxley_diams = np.sort(jaxley_diams)

    for sec in h.allsec():
        h.delete_section(sec=sec)

    cell = h.Import3d_SWC_read()
    cell.input(fname)
    i3d = h.Import3d_GUI(cell, False)
    i3d.instantiate(None)

    neuron_diams = []
    for sec in h.allsec():
        sec.nseg = ncomp
        for seg in sec:
            neuron_diams.append(seg.diam)
    neuron_diams = np.asarray(neuron_diams)
    neuron_diams = np.sort(neuron_diams)

    max_error = np.max(np.abs(jaxley_diams - neuron_diams))
    assert max_error < 0.5, f"radiuses do not match, error {max_error}."


@pytest.mark.parametrize("reader_backend", ["graph"])
@pytest.mark.parametrize(
    "file",
    [
        "morph_3_types.swc",
        "morph_3_types_single_point_soma.swc",
        "morph_ca1_n120.swc",
        "morph_l5pc_with_axon.swc",
    ],
)
def test_swc_types(reader_backend, file):
    # Can not use full morphology because of branch sorting.
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", file)
    backend_kwargs = (
        {"ignore_swc_tracing_interruptions": False} if reader_backend == "graph" else {}
    )
    cell = jx.read_swc(fname, ncomp=1, backend=reader_backend, **backend_kwargs)

    # First iteration is with default `ncomp`. At the end of the first loop we change
    # ncomp with `set_ncomp`
    for i in range(2):
        desired_numbers_of_comps = {
            "morph_3_types.swc": {"soma": 1, "axon": 1, "basal": 1},
            "morph_3_types_single_point_soma.swc": {
                "soma": 1,
                "axon": 1,
                "basal": 1,
            },
            "morph_ca1_n120.swc": {"soma": 2, "basal": 101, "apical": 53},
            "morph_l5pc_with_axon.swc": {
                "soma": 1,
                "axon": 128,
                "basal": 66,
                "apical": 129,
            },
        }
        # Test soma.
        for key, n_desired in desired_numbers_of_comps[file].items():
            if i == 1 and key in ["soma", "basal"]:
                n_desired += 2  # After `set_ncomp` we should have two more comps.
            n_comps_in_morph = len(cell.__getattr__(key).nodes)
            assert (
                n_comps_in_morph == n_desired
            ), f"{key} has {n_comps_in_morph} != {n_desired} comps!"

        # Additional tests to ensure that `groups` get updated appropriately.
        cell.soma.branch(0).set_ncomp(3)
        cell.basal.branch(0).set_ncomp(3)


def test_single_branch_swc():
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", "morph_single_branch.swc")
    cell = jx.read_swc(fname, ncomp=1)
    cell.branch(0).set_ncomp(3)
    cell.set_ncomp(4)
    cell[0, 0].record()
    v = jx.integrate(cell, t_max=1.0)
    assert np.invert(np.any(np.isnan(v))), "Found a NaN."



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>
from abc import ABC, abstractmethod
from typing import Dict, Optional, Tuple
from warnings import warn

from jax import Array
from jax.typing import ArrayLike


class Channel:
    """Channel base class. All channels inherit from this class.

    A channel in Jaxley is everything that modifies the membrane voltage via its
    current returned by the `compute_current()` method.

    As in NEURON, a `Channel` is considered a distributed process, which means that its
    conductances are to be specified in `S/cm2` and its currents are to be specified in
    `uA/cm2`."""

    _name = None
    channel_params = None
    channel_states = None
    current_name = None

    def __init__(self, name: Optional[str] = None):
        contact = (
            "If you have any questions, please reach out via email to "
            "michael.deistler@uni-tuebingen.de or create an issue on Github: "
            "https://github.com/jaxleyverse/jaxley/issues. Thank you!"
        )
        if (
            not hasattr(self, "current_is_in_mA_per_cm2")
            or not self.current_is_in_mA_per_cm2
        ):
            raise ValueError(
                "The channel you are using is deprecated. "
                "In Jaxley version 0.5.0, we changed the unit of the current returned "
                "by `compute_current` of channels from `uA/cm^2` to `mA/cm^2`. Please "
                "update your channel model (by dividing the resulting current by 1000) "
                "and set `self.current_is_in_mA_per_cm2=True` as the first line "
                f"in the `__init__()` method of your channel. {contact}"
            )

        self._name = name if name else self.__class__.__name__

    @property
    def name(self) -> Optional[str]:
        """The name of the channel (by default, this is the class name)."""
        return self._name

    def change_name(self, new_name: str):
        """Change the channel name.

        Args:
            new_name: The new name of the channel.

        Returns:
            Renamed channel, such that this function is chainable.
        """
        old_prefix = self._name + "_"
        new_prefix = new_name + "_"

        self._name = new_name
        self.channel_params = {
            (
                new_prefix + key[len(old_prefix) :]
                if key.startswith(old_prefix)
                else key
            ): value
            for key, value in self.channel_params.items()
        }

        self.channel_states = {
            (
                new_prefix + key[len(old_prefix) :]
                if key.startswith(old_prefix)
                else key
            ): value
            for key, value in self.channel_states.items()
        }
        return self

    def update_states(self, states, dt, v, params) -> tuple[Array, tuple[Array, Array]]:
        """Return the updated states."""
        raise NotImplementedError

    def compute_current(
        self, states: dict[str, Array], v: float, params: dict[str, Array]
    ):
        """Given channel states and voltage, return the current through the channel.

        Args:
            states: All states of the compartment.
            v: Voltage of the compartment in mV.
            params: Parameters of the channel (conductances in `S/cm2`).

        Returns:
            Current in `uA/cm2`.
        """
        raise NotImplementedError

    def init_state(
        self,
        states: dict[str, ArrayLike],
        v: ArrayLike,
        params: dict[str, ArrayLike],
        delta_t: float,
    ):
        """Initialize states of channel."""
        return {}



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import os
from copy import deepcopy

import jax

from jaxley.connect import connect

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

import matplotlib.pyplot as plt
import numpy as np
import pytest

import jaxley as jx
from jaxley.synapses import IonotropicSynapse


def test_cell(SimpleMorphCell):
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", "morph_ca1_n120.swc")
    cell = SimpleMorphCell(fname, ncomp=1)
    cell.branch(0).set_ncomp(2)  # test inhomogeneous ncomp

    # Plot 1.
    _, ax = plt.subplots(1, 1, figsize=(3, 3))
    ax = cell.vis(ax=ax)
    ax = cell.branch([0, 1, 2]).vis(ax=ax, color="r")
    ax = cell.branch(1).loc(0.9).vis(ax=ax, color="b")

    # Plot 2.
    cell.branch(0).add_to_group("soma")
    cell.branch(1).add_to_group("soma")
    ax = cell.soma.vis()


def test_network(SimpleMorphCell):
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", "morph_ca1_n120.swc")
    cell1 = SimpleMorphCell(fname, ncomp=1)
    cell2 = SimpleMorphCell(fname, ncomp=1)
    cell3 = SimpleMorphCell(fname, ncomp=1)

    net = jx.Network([cell1, cell2, cell3])
    connect(
        net.cell(0).branch(0).loc(0.0),
        net.cell(1).branch(0).loc(0.0),
        IonotropicSynapse(),
    )
    connect(
        net.cell(0).branch(0).loc(0.0),
        net.cell(2).branch(0).loc(0.0),
        IonotropicSynapse(),
    )

    # Plot 1.
    _, ax = plt.subplots(1, 1, figsize=(3, 3))
    ax = net.cell([0, 1]).vis(ax=ax)
    ax = net.cell(2).vis(ax=ax, color="r", type="line")
    ax = net.cell(2).vis(ax=ax, color="r", type="scatter")
    ax = net.cell(0).branch(np.arange(10).tolist()).vis(ax=ax, color="b")

    # Plot 2.
    ax = net.vis(detail="full", type="line")
    ax = net.vis(detail="full", type="scatter")

    # Plot 3.
    net.vis(detail="point")

    # Plot 4.
    net.arrange_in_layers([2, 1])
    net.vis(detail="point")

    # Plot 5.
    net.arrange_in_layers([2, 1])
    net.vis(detail="full")

    # Plot 5.
    net.cell(0).add_to_group("excitatory")
    net.cell(1).add_to_group("excitatory")
    ax = net.excitatory.vis()


def test_vis_networks_built_from_scratch(SimpleComp, SimpleBranch, SimpleCell):
    comp = SimpleComp(copy=True)
    branch = SimpleBranch(4)
    cell = SimpleCell(5, 3)
    cell.branch(0).set_ncomp(3)  # test inhomogeneous ncomp

    net = jx.Network([cell, cell])
    connect(
        net.cell(0).branch(0).loc(0.0),
        net.cell(1).branch(0).loc(0.0),
        IonotropicSynapse(),
    )
    connect(
        net.cell(0).branch(0).loc(0.0),
        net.cell(1).branch(2).loc(0.0),
        IonotropicSynapse(),
    )
    net.compute_xyz()

    # Plot 1.
    _, ax = plt.subplots(1, 1, figsize=(3, 3))
    ax = net.vis(detail="full", ax=ax)

    # Plot 2.
    _, ax = plt.subplots(1, 1, figsize=(3, 3))
    net.cell(0).move(0, 100)
    ax = net.vis(detail="full", ax=ax)

    # Plot 3.
    _, ax = plt.subplots(1, 1, figsize=(3, 3))
    comp.compute_xyz()
    ax = comp.vis(ax=ax)

    # Plot 4.
    _, ax = plt.subplots(1, 1, figsize=(3, 3))
    branch.compute_xyz()
    ax = branch.vis(ax=ax)


def test_mixed_network(SimpleMorphCell):
    dirname = os.path.dirname(__file__)
    fname = os.path.join(dirname, "swc_files", "morph_ca1_n120.swc")
    cell1 = SimpleMorphCell(fname, ncomp=1)

    comp = jx.Compartment()
    branch = jx.Branch(comp, 4)
    cell2 = jx.Cell(branch, parents=[-1, 0, 0, 1, 1])

    net = jx.Network([cell1, cell2])
    connect(
        net.cell(0).branch(0).loc(0.0),
        net.cell(1).branch(0).loc(0.0),
        IonotropicSynapse(),
    )
    connect(
        net.cell(0).branch(0).loc(0.0),
        net.cell(1).branch(1).loc(0.0),
        IonotropicSynapse(),
    )

    net.compute_xyz()
    net.cell(0).move(0, 800)
    net.cell(1).move(0, -800)
    net.rotate(180)

    before_xyzrs = deepcopy(net.xyzr[len(cell1.xyzr) :])
    net.cell(1).rotate(90)
    after_xyzrs = net.xyzr[len(cell1.xyzr) :]
    # Test that rotation worked as expected.
    for b, a in zip(before_xyzrs, after_xyzrs):
        assert np.allclose(b[:, 0], -a[:, 1], atol=1e-6)
        assert np.allclose(b[:, 1], a[:, 0], atol=1e-6)

    _ = net.vis(detail="full")


def test_volume_plotting(
    SimpleComp, SimpleBranch, SimpleCell, SimpleNet, SimpleMorphCell
):
    comp = SimpleComp()
    branch = SimpleBranch(2)
    cell = SimpleCell(2, 2)
    cell.branch(0).set_ncomp(3)  # test inhomogeneous ncomp
    net = SimpleNet(2, 2, 2)

    for module in [comp, branch, cell, net]:
        module.compute_xyz()

    fname = os.path.join(
        os.path.dirname(__file__), "swc_files", "morph_ca1_n120_250.swc"
    )
    morph_cell = SimpleMorphCell(fname, ncomp=1)

    fig, ax = plt.subplots()
    for module in [comp, branch, cell, morph_cell]:
        module.vis(type="comp", ax=ax, resolution=6)
    net.vis(type="comp", ax=ax, cell_plot_kwargs={"resolution": 6})
    plt.close(fig)

    # test 3D plotting
    for module in [comp, branch, cell, morph_cell]:
        module.vis(type="comp", dims=[0, 1, 2], resolution=6)
    net.vis(type="comp", dims=[0, 1, 2], cell_plot_kwargs={"resolution": 6})
    plt.close()

    # test morph plotting (does not work if no radii in xyzr)
    morph_cell.branch(1).vis(type="morph")
    morph_cell.branch(1).vis(
        type="morph", dims=[0, 1, 2], resolution=6
    )  # plotting whole thing takes too long
    plt.close()

    # In #725, we had discovered a bug when `ncomp=2`. This test ensures that this
    # bug does not resurface.
    morph_cell = SimpleMorphCell(fname, ncomp=2)
    morph_cell.vis(type="morph", dims=[0, 1], linewidth=0.0, resolution=6)
    plt.close()



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax.numpy as jnp


def step_current(
    i_delay: float,
    i_dur: float,
    i_amp: float,
    delta_t: float,
    t_max: float,
    i_offset: float = 0.0,
):
    """
    Return step current in unit nA.

    Unlike the `datapoint_to_step()` method, this takes a single value for the amplitude
    and returns a single step current. The output of this function can be passed to
    `.stimulate()`, but not to `integrate(..., currents=)`.
    """
    dt = delta_t
    window_start = int(i_delay / dt)
    window_end = int((i_delay + i_dur) / dt)
    time_steps = int(t_max // dt) + 2
    current = jnp.zeros((time_steps,)) + i_offset
    return current.at[window_start:window_end].set(i_amp)


def datapoint_to_step_currents(
    i_delay: float,
    i_dur: float,
    i_amp: jnp.asarray,
    delta_t: float,
    t_max: float,
    i_offset: float = 0.0,
):
    """
    Return several step currents in unit nA.

    Unlike the `step_current()` method, this takes a vector of amplitude and returns
    a step current for each value in the vector. The output of this function can be
    passed to `integrate(..., currents=)`, but can not be passed to `.stimulate()`.
    """
    dim = len(i_amp)
    dt = delta_t
    window_start = int(i_delay / dt)
    window_end = int((i_delay + i_dur) / dt)
    time_steps = int(t_max // dt) + 2
    current = jnp.zeros((time_steps, dim)) + i_offset
    return current.at[window_start:window_end, :].set(i_amp).T



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")


import jax.numpy as jnp
import numpy as np
import pytest
from jax import jit, value_and_grad

import jaxley as jx
from jaxley.channels import HH


def _run_long_branch(dt, t_max, current, branch):
    branch.insert(HH())

    branch.loc("all").make_trainable("radius", 1.0)
    params = branch.get_parameters()

    branch.loc(0.0).record()
    branch.loc(0.0).stimulate(current)

    def loss(params):
        s = jx.integrate(branch, params=params)
        return s[0, -1]

    jitted_loss_grad = jit(value_and_grad(loss))

    l, g = jitted_loss_grad(params)
    return l, g


def _run_short_branches(dt, t_max, current, cell):
    cell.insert(HH())

    cell.branch("all").loc("all").make_trainable("radius", 1.0)
    params = cell.get_parameters()

    cell.branch(0).loc(0.0).record()
    cell.branch(0).loc(0.0).stimulate(current)

    def loss(params):
        s = jx.integrate(cell, params=params)
        return s[0, -1]

    jitted_loss_grad = jit(value_and_grad(loss))

    l, g = jitted_loss_grad(params)
    return l, g


@pytest.mark.slow
def test_equivalence(SimpleBranch, SimpleCell):
    """Test whether a single long branch matches a cell of two shorter branches."""
    dt = 0.025
    t_max = 5.0  # ms
    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
    )
    l1, g1 = _run_long_branch(dt, t_max, current, SimpleBranch(8))
    l2, g2 = _run_short_branches(dt, t_max, current, SimpleCell(2, 4))

    assert np.allclose(l1, l2), "Losses do not match."

    for g_1, g_2 in zip(g1, g2):
        for key in g_1:
            assert np.allclose(g_1[key], g_2[key]), "Gradients do not match."



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

# Non-capacitive channels are all channels that do not follow the default
# Hodgkin-Huxley update:
#
# 1/C dV/dt = i
#
# These non-capacitive achieve this by modifying the voltage as part of their
# `update_states()` method.



# This file is part of Jaxley, a differentiable neuroscience simulator. Jaxley is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

import jax

jax.config.update("jax_enable_x64", True)
jax.config.update("jax_platform_name", "cpu")

import jax.numpy as jnp

import jaxley as jx
from jaxley.channels import HH


def test_compose_branch():
    """Test inserting to comp and composing to branch equals inserting to branch."""
    dt = 0.025
    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
    )

    comp1 = jx.Compartment()
    comp1.insert(HH())
    comp2 = jx.Compartment()
    branch1 = jx.Branch([comp1, comp2])
    branch1.loc(0.0).record()
    branch1.loc(0.0).stimulate(current)

    comp = jx.Compartment()
    branch2 = jx.Branch(comp, ncomp=2)
    branch2.loc(0.0).insert(HH())
    branch2.loc(0.0).record()
    branch2.loc(0.0).stimulate(current)

    voltages1 = jx.integrate(branch1, delta_t=dt)
    voltages2 = jx.integrate(branch2, delta_t=dt)

    assert jnp.max(jnp.abs(voltages1 - voltages2)) < 1e-8


def test_compose_cell():
    """Test inserting to branch and composing to cell equals inserting to cell."""
    ncomp_per_branch = 4
    dt = 0.025
    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
    )

    comp = jx.Compartment()

    branch1 = jx.Branch(comp, ncomp_per_branch)
    branch1.insert(HH())
    branch2 = jx.Branch(comp, ncomp_per_branch)
    cell1 = jx.Cell([branch1, branch2], parents=[-1, 0])
    cell1.branch(0).loc(0.0).record()
    cell1.branch(0).loc(0.0).stimulate(current)

    branch = jx.Branch(comp, ncomp_per_branch)
    cell2 = jx.Cell(branch, parents=[-1, 0])
    cell2.branch(0).insert(HH())
    cell2.branch(0).loc(0.0).record()
    cell2.branch(0).loc(0.0).stimulate(current)

    voltages1 = jx.integrate(cell1, delta_t=dt)
    voltages2 = jx.integrate(cell2, delta_t=dt)

    assert jnp.max(jnp.abs(voltages1 - voltages2)) < 1e-8


def test_compose_net():
    """Test inserting to cell and composing to net equals inserting to net."""
    ncomp_per_branch = 4
    dt = 0.025
    current = jx.step_current(
        i_delay=0.5, i_dur=1.0, i_amp=0.1, delta_t=0.025, t_max=5.0
    )

    comp = jx.Compartment()
    branch = jx.Branch(comp, ncomp_per_branch)

    cell1 = jx.Cell(branch, parents=[-1, 0, 0])
    cell1.insert(HH())
    cell2 = jx.Cell(branch, parents=[-1, 0, 0])
    net1 = jx.Network([cell1, cell2])
    net1.cell(0).branch(0).loc(0.0).record()
    net1.cell(0).branch(0).loc(0.0).stimulate(current)

    cell = jx.Cell(branch, parents=[-1, 0, 0])
    net2 = jx.Network([cell, cell])
    net2.cell(0).insert(HH())
    net2.cell(0).branch(0).loc(0.0).record()
    net2.cell(0).branch(0).loc(0.0).stimulate(current)

    voltages1 = jx.integrate(net1, delta_t=dt)
    voltages2 = jx.integrate(net2, delta_t=dt)

    assert jnp.max(jnp.abs(voltages1 - voltages2)) < 1e-8
